/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: cattach_SC.c,v 1.2 2001/01/03 09:55:55 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "config.h"
#include "global.h"
#include "smartcard_functions.h"
#include "cfs_SC_functions.h"

// cattach_SC is a wrapper for the cfs function cattach in order to
// use smartcards to store the passphrase.
// It also uses two files; the first to registred the currently
// attached directories; the second to put in correspondence
// ciphered directories and their clear name with a supplementary
// flag to indicate if the user want its directories be auto attached.
//
// So, this wrapper uses and manages these files, allows to attach
// directories whose auto attached flag is set to auto, and finally calls 
// the cattach binary.
//
// This binary takes three options:
//    * -q: the program will not print any messages.
//    * -a: directories of the user cfs fstab file whose auto attached 
//          flag is set to auto will be attached.
//    * -p pin: the user pin code is given in parameter. This feature 
//              should be reserved to the cfs PAM module...
//
// If the -a option is not set, user must provide the ciphered dir path 
// and the attaching clear name. He can just provide the ciphered dir
// path, in this case the corresponding attaching clear name will be
// seeked int the user cdfs fstab file. 
/////////////////////////////////////////////////////////////////////////

// Internal functions
/////////////////////

// Print program usage informations
void usage(char *progName);

// Working part of the program...
int cattach_SC(char *clearDirName, char *cipheredDirName,
	unsigned char *cfsSecret, int quiet);

int main(int argc, char **argv)
{
	int found;
	int quiet = 0;				// Function is verbose by default
	int rCode, result;
	int attachAll = 0;
	int pinGiven = 0;
	int failureDuringAttaching = 0;
	int noClearDirNameGiven = 0;
	char *pinReaded;
	char pin[USR_PIN_SIZE + 1];
	char clearDirName[CFS_DIR_NAME_SIZE];
	char cipheredDirName[CFS_DIR_NAME_SIZE];
	unsigned char cfsSecret[CFS_SECRET_SIZE + 1];

	// Get the arguments
	////////////////////

	while (1)
	{
		rCode = getopt(argc, argv, "p:aq");

		if (rCode == -1)
			break;

		switch (rCode)
		{
		case 'p':				// The pin is given as an argument
			pinGiven = 1;
			strncpy(pin, optarg, USR_PIN_SIZE);
			pin[USR_PIN_SIZE] = '\0';
			break;

		case 'a':
			attachAll = 1;		// Attach all option set
			break;

		case 'q':				// Be silent
			quiet = 1;
			break;

		case ':':				// Wrong argument
			if (!quiet)
				usage(argv[0]);
			break;

		case '?':				// Wrong argument
			if (!quiet)
				usage(argv[0]);
			break;
		}
	}

	if (attachAll && (optind < argc))
		if (!quiet)
			usage(argv[0]);		// Too much arguments

	if (!attachAll)
	{
		if (optind < argc)
		{
			// Fist argument is the ciphered directory name
			strncpy(cipheredDirName, argv[optind], CFS_DIR_NAME_SIZE);
			cipheredDirName[strlen(argv[optind])] = '\0';

			if (argv[optind + 1] != NULL)
			{
				// Second (optional) argument is the attachement clear name
				strncpy(clearDirName, argv[optind + 1], CFS_DIR_NAME_SIZE);
				clearDirName[strlen(argv[optind + 1])] = '\0';

				if (argv[optind + 2] != NULL)	// Too much argument
					if (!quiet)
						usage(argv[0]);
			}
			else
				noClearDirNameGiven = 1;
		}
		else if (!quiet)
			usage(argv[0]);
	}

	if (!pinGiven)
	{
		// No PIN given, we have to ask for the PIN to the user
		pinReaded = getpass("Enter the your pin: ");
		strncpy(pin, pinReaded, USR_PIN_SIZE);
		pin[USR_PIN_SIZE] = '\0';
	}

	// Try to get the CFS secret stored in the card
	rCode = get_cfs_secret(pin, cfsSecret, quiet);
	cfsSecret[CFS_SECRET_SIZE] = '\n';

	if (rCode == ERROR_CARD_NOT_FORMATED)
	{
		fprintf(stderr, "ERROR: this CARD seems to be NOT FORMATTED!\n");
		exit(EXIT_FAILURE);
	}

	if (rCode == ERROR_VERIFY_PIN)
	{
		// Specific wrong PIN error 
		if (!quiet)
			fprintf(stderr, "ERROR: user PIN code is WRONG!\n");
		exit(EXIT_FAILURE);
	}

	if (rCode == EXIT_FAILURE)
	{
		// Other error, can't get the CFS secret
		if (!quiet)
			fprintf(stderr, "UNABLE to get the CFS secret from the card\n");
		exit(EXIT_FAILURE);
	}

	// Create the user cfs mtab database if it doesn't still exist
	rCode = mtab_create(quiet);

	if (rCode == EXIT_FAILURE)
	{
		// Can't create the database
		if (!quiet)
			fprintf(stderr,
				"UNABLE to create the the user cfs mtab database\n");
		exit(EXIT_FAILURE);
	}

	if (attachAll)
	{
		// Attach all the directories registred in the user cfs fstab 
		// file whose auto attach flag is set to auto
		/////////////////////////////////////////////////////////////

		int lineNr = 1;			// References the number of the line whose flag is set to auto
		int found = 0;			// Set to 1 if an candidate directory is available

		// Get the first line. We want to know the ciphered directory name 
		// and the attaching clear name.
		rCode = give_a_cfs_fstab_entry(cipheredDirName, clearDirName,
			lineNr, &found, quiet);
		if (rCode == EXIT_FAILURE)
		{
			if (!quiet)
				fprintf(stderr,
					"UNABLE to get an entry from the cfs fstab file\n");
			failureDuringAttaching = 1;
			goto OUT;
		}

		if (!found)				// No line found...
			if (!quiet)
				printf("Nothing to attach\n");

		while (found)
		{
			if (!quiet)
				fprintf(stderr, "Attaching %s to %s\n", cipheredDirName,
					clearDirName);

			// Attach this directory to the clear name 
			rCode =
				cattach_SC(clearDirName, cipheredDirName, cfsSecret, quiet);

			if (rCode == EXIT_FAILURE)
			{
				failureDuringAttaching = 1;
				// We don't return, other attachement should be tried.
			}

			// Get the next line
			rCode = give_a_cfs_fstab_entry(cipheredDirName, clearDirName,
				++lineNr, &found, quiet);
			if (rCode == EXIT_FAILURE)
			{
				if (!quiet)
					fprintf(stderr,
						"UNABLE to get an entry from the cfs fstab file\n");
				failureDuringAttaching = 1;
				goto OUT;
			}
		}
	}

	else if (noClearDirNameGiven)
	{
		// Get the clear name stored in the cfs fstab file
		// in correspondence with the given ciphered directory name
		///////////////////////////////////////////////////////////

		rCode =
			get_a_cfs_fstab_clear_name(cipheredDirName, clearDirName, &found,
			quiet);

		if (rCode == EXIT_FAILURE)
		{
			if (!quiet)
				fprintf(stderr,
					"UNABLE to get informations from the cfs fstab file\n");
			failureDuringAttaching = 1;
			goto OUT;
		}

		if (!found)
		{
			if (!quiet)
				fprintf(stderr,
					"Error, this ciphered directory %s is not registred in the cfs fstab file\n",
					cipheredDirName);
			failureDuringAttaching = 1;
			goto OUT;
		}

		// Call the cattach_SC function
		rCode = cattach_SC(clearDirName, cipheredDirName, cfsSecret, quiet);

		if (rCode == EXIT_FAILURE)
		{
			failureDuringAttaching = 1;
			goto OUT;
		}
	}

	else
	{
		// We have all the arguments, just call the cattach_SC function
		rCode = cattach_SC(clearDirName, cipheredDirName, cfsSecret, quiet);

		if (rCode == EXIT_FAILURE)
		{
			failureDuringAttaching = 1;
			goto OUT;
		}
	}

  OUT:							// was there a failure during the attaching operation?
	if (failureDuringAttaching)
	{
		// Perhaps database is always empty ?
		rCode = is_mtab_empty(&result, quiet);

		if (rCode == EXIT_FAILURE)
		{
			if (!quiet)
				fprintf(stderr, "UNABLE to use the user cfs mtab database\n");
			exit(EXIT_FAILURE);
		}

		if (result)
		{
			// Yes, the user cfs mtab database is empty, so we can delete it
			rCode = mtab_delete(quiet);

			if (rCode == EXIT_FAILURE)
			{
				if (!quiet)
					fprintf(stderr,
						"UNABLE to remove the user cfs mtab database\n");
				exit(EXIT_FAILURE);
			}
		}

		exit(EXIT_FAILURE);
	}

	exit(EXIT_SUCCESS);
}

int cattach_SC(char *clearDirName, char *cipheredDirName,
	unsigned char *cfsSecret, int quiet)
{
	int result;
	int fildes[2];
	int pid, rCode, status;

	// Verify that this directory is not already attached
	// It mustn't be registred in the user cfs mtab database
	rCode = is_dir_in_mtab(clearDirName, &result, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to use the user cfs mtab database\n");
		return (EXIT_FAILURE);
	}

	if (result)
	{
		if (!quiet)
			fprintf(stderr, "The directory %s is already attached to %s\n",
				cipheredDirName, clearDirName);
		// It's not a failure, but perhaps it's not really a success:)
		return (EXIT_SUCCESS);
	}

	// Create a pipe to directly send the CFS secret to cattach
	rCode = pipe(fildes);

	if (rCode == -1)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to create the pipe: %s\n",
				strerror(errno));
		return (EXIT_FAILURE);
	}
	// Now, we have to fork before call the binary cattach
	pid = fork();

	if (pid == -1)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to fork: %s\n", strerror(errno));
		return (EXIT_FAILURE);
	}

	if (pid == 0)
	{
		// We are into the child process
		//
		// Prepare the environment before the call of the cattach binary
		////////////////////////////////////////////////////////////////

		// Dup stdin to read descriptor
		rCode = dup2(fildes[0], 0);

		if (rCode == -1)
		{
			if (!quiet)
				fprintf(stderr, "UNABLE to dup stdin on the pipe: %s\n",
					strerror(errno));
			exit(EXIT_FAILURE);
		}

		// If quiet is set, close stdout and stderr
		// Good method to silence the process!!
		if (quiet)
		{
			fclose(stdout);
			fclose(stderr);
		}

		// Execute cattach binary
		// cattach(1): The -- option will read the key from standard input
		// So cattach will read the key from the pipe on which the father
		// will send the key...
		//////////////////////////////////////////////////////////////////

		rCode = execlp((CFS_BIN_DIR CATTACH), CATTACH, "--",
			cipheredDirName, clearDirName, NULL);

		if (rCode == -1)
		{
			if (!quiet)
				fprintf(stderr, "UNABLE to execute %s: %s\n", CATTACH,
					strerror(errno));
			exit(EXIT_FAILURE);
		}
	}

	else
	{
		// We are still into the father process
		///////////////////////////////////////

		// Write the CFS secret on the pipe
		rCode = write(fildes[1], cfsSecret, CFS_SECRET_SIZE + 1);

		if (rCode != CFS_SECRET_SIZE + 1)
		{
			if (!quiet)
				fprintf(stderr,
					"UNABLE to write the CFS secret on pipe: %s\n",
					strerror(errno));
			return (EXIT_FAILURE);
		}

		// Wait for the end of the child
		rCode = waitpid(pid, &status, 0);

		if (WEXITSTATUS(status) != 0)
		{
			// Child doesn't exit normally, so directory isn't attached

			if (!quiet)
				fprintf(stderr, "UNABLE to attach the directory %s to %s\n",
					cipheredDirName, clearDirName);
			return (EXIT_FAILURE);
		}

		// Update cfs mtab to add this newly attached directory  
		rCode = add_dir_in_mtab(clearDirName, cipheredDirName, quiet);

		if (rCode == EXIT_FAILURE)
		{
			if (!quiet)
				fprintf(stderr,
					"UNABLE to add the directory %s in cfs_SC_mtab\n",
					clearDirName);
			return (EXIT_FAILURE);
		}
	}

	return (EXIT_SUCCESS);
}

void usage(char *progName)
{
	// Print program usage informations
	///////////////////////////////////

	fprintf(stderr, "Usage: %s [-p pin] -a | directory [name]\n", progName);
	exit(EXIT_FAILURE);
}
