/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: cfs_SC_functions.c,v 1.2 2001/01/03 09:55:55 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <gdbm.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <pwd.h>

#include "config.h"
#include "global.h"
#include "cfs_SC_functions.h"

// Internal functions
/////////////////////

// Returns the name of the user cfs mtab database in dbName.
int get_db_name(char *dbName, int quiet);

// Returns the name of the user cfs fstab file in fileName.
int get_file_name(char *fileName, int quiet);

// Returns the absolute path of relativePath in absolutePath.
int get_absolute_path(char *relativePath, char *absolutePath, int quiet);

int is_dir_in_mtab(char *clearDirName, int *result, int quiet)
{
	// This functions sets result to 1 if the clearDirName clear directory
	// is registred in the user cfs mtab database. Sets result to 0 else.
	//////////////////////////////////////////////////////////////////////

	int rCode;
	char dbName[MAX_DBNAME_SIZE];
	datum keyToSearch;
	GDBM_FILE dbf;

	// Get the user cfs mtab database name
	rCode = get_db_name(dbName, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to get the user cfs mtab database name\n");
		return (EXIT_FAILURE);
	}

	// Search the clear directory name in this database
	///////////////////////////////////////////////////

	// Prepare the key  
	keyToSearch.dptr = clearDirName;
	keyToSearch.dsize = strlen(clearDirName);

	// Open the database
	dbf = gdbm_open(dbName, DB_BLOC_SIZE, GDBM_READER, 0, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to open the user cfs mtab database (%s): %s\n",
				dbName, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Is the key in the database
	// An eventually error will be directly displayed 
	// by the gdbm_exists function
	if (gdbm_exists(dbf, keyToSearch))
		*result = 1;			// Yes
	else
		*result = 0;			// No

	gdbm_close(dbf);

	return (EXIT_SUCCESS);
}

int add_dir_in_mtab(char *clearDirName, char *cipheredDirName, int quiet)
{
	// This function adds clearDirName to the user cfs mtab database.
	// We also store as data the corresponding cipheredDirName which
	// could be used furthermore.
	/////////////////////////////////////////////////////////////////

	int rCode;
	datum keyToAdd, dataToAdd;
	char dbName[MAX_DBNAME_SIZE];
	GDBM_FILE dbf;

	// Get the user cfs mtab database name
	rCode = get_db_name(dbName, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to get the user cfs mtab database name\n");
		return (EXIT_FAILURE);
	}

	// Add the key to the database
	// Return an error if the directory is already registred on the database
	////////////////////////////////////////////////////////////////////////

	// Prepare the key
	keyToAdd.dptr = clearDirName;
	keyToAdd.dsize = strlen(clearDirName);

	// Prepare the data
	dataToAdd.dptr = cipheredDirName;
	dataToAdd.dsize = strlen(cipheredDirName);

	// Open the user cfs mtab database
	dbf = gdbm_open(dbName, DB_BLOC_SIZE, GDBM_WRITER, 0, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to open the user cfs mtab database (%s): %s\n",
				dbName, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Store the key and the data
	// An eventually error will be directly displayed 
	// by the gdbm_store function
	rCode = gdbm_store(dbf, keyToAdd, dataToAdd, GDBM_INSERT);

	gdbm_close(dbf);

	if (rCode != 0)
		return (EXIT_FAILURE);	// Unable to store the key
	else
		return (EXIT_SUCCESS);	// Key stored
}

int delete_dir_in_mtab(char *clearDirName, int quiet)
{
	// This function remove thes clearDirName from the user cfs 
	// mtab database. An error is returned if the key is not 
	// registred in the database.
	///////////////////////////////////////////////////////////

	int rCode;
	datum keyToDelete;
	char dbName[MAX_DBNAME_SIZE];
	GDBM_FILE dbf;

	// Get the use cfs mtab database name
	rCode = get_db_name(dbName, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to get the user cfs mtab database name\n");
		return (EXIT_FAILURE);
	}

	// Prepare the key
	keyToDelete.dptr = clearDirName;
	keyToDelete.dsize = strlen(clearDirName);

	// Open the user cfs mtab database  
	dbf = gdbm_open(dbName, DB_BLOC_SIZE, GDBM_WRITER, 0, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to open the user cfs mtab database (%s): %s\n",
				dbName, strerror(errno));
		gdbm_close(dbf);
		return (EXIT_FAILURE);
	}

	// Remove the key
	// An eventually error will be directly displayed 
	// by the gdbm_delete function
	rCode = gdbm_delete(dbf, keyToDelete);

	gdbm_close(dbf);

	if (rCode == -1)
		return (EXIT_FAILURE);	// Unable to remove the key
	else
		return (EXIT_SUCCESS);	// Key removed
}

int is_mtab_empty(int *result, int quiet)
{
	// This function set result to 1 is the user cfs mtab database is
	// empty; set to 0 else.
	/////////////////////////////////////////////////////////////////

	int rCode;
	char dbName[MAX_DBNAME_SIZE];
	datum keyToSearch;
	GDBM_FILE dbf;

	// Get the user cfs mtab database name
	rCode = get_db_name(dbName, quiet);

	if (rCode == EXIT_FAILURE)
	{
		fprintf(stderr, "UNABLE to get the user cfs mtab database name\n");
		return (EXIT_FAILURE);
	}

	// Open the user cfs mtab database
	dbf = gdbm_open(dbName, DB_BLOC_SIZE, GDBM_WRITER, 0, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		if (errno == ENOENT)
		{
			// The database doesn't exist, we consider that it's empty.
			*result = 1;
			return (EXIT_SUCCESS);
		}
		else
		{
			// Other probem, it's an error
			if (!quiet)
				fprintf(stderr,
					"UNABLE to open the user cfs mtab database (%s): %s\n",
					dbName, strerror(errno));
			return (EXIT_FAILURE);
		}
	}

	// Try to get the first key of the database.
	// An eventually error will be directly displayed 
	// by the gdbm_firstkey function
	keyToSearch = gdbm_firstkey(dbf);

	gdbm_close(dbf);

	if (keyToSearch.dptr == NULL)
		*result = 1;			// Such key doesn't exist: database is empty
	else
		*result = 0;			// Such key exist: database is not empty

	return (EXIT_SUCCESS);
}

int give_dir_in_mtab(char *clearDirName, int *result, int quiet)
{
	// This functions return a clearDirName registred in the user 
	// cfs mtab database an set result to 1.
	// If the database is empty result is set to 0.
	/////////////////////////////////////////////////////////////

	int rCode;
	char dbName[MAX_DBNAME_SIZE];
	datum keyToSearch;
	GDBM_FILE dbf;

	// Get the user cfs mtab database name
	rCode = get_db_name(dbName, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to get the user cfs mtab database name\n");
		return (EXIT_FAILURE);
	}

	// Open the user cfs mtab database
	dbf = gdbm_open(dbName, DB_BLOC_SIZE, GDBM_WRITER, 0, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		if (errno == ENOENT)
		{
			// The database doesn't exit, it's not an error but
			// we indicate that we haven't any clear name available. 
			*result = 0;
			return (EXIT_SUCCESS);
		}
		else
		{
			// This other problem is considered as an error
			if (!quiet)
				fprintf(stderr,
					"UABLE to open the user cfs mtab database (%s): %s\n",
					dbName, strerror(errno));
			return (EXIT_FAILURE);
		}
	}

	// Try to get the first key of the database
	// An eventually error will be directly displayed 
	// by the gdbm_firstkey function
	keyToSearch = gdbm_firstkey(dbf);

	if (keyToSearch.dptr != NULL)
	{
		// We have found a key, copy it's value in clearDirName
		strncpy(clearDirName, keyToSearch.dptr, keyToSearch.dsize);
		clearDirName[keyToSearch.dsize] = '\0';
		*result = 1;			// Key found
	}
	else
		*result = 0;			// No key found

	gdbm_close(dbf);

	return (EXIT_SUCCESS);
}

int mtab_create(int quiet)
{
	// This function creates the user cfs mtb database.
	// It returns an error if the database already exists.
	//////////////////////////////////////////////////////

	int rCode;
	char dbName[MAX_DBNAME_SIZE];
	GDBM_FILE dbf;

	// Get the user cfs mtab database name
	rCode = get_db_name(dbName, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to get the user cfs mtab database name\n");
		return (EXIT_FAILURE);
	}

	// Create the database
	dbf = gdbm_open(dbName, DB_BLOC_SIZE, GDBM_WRCREAT,
		OPEN_CREATE_DB_RIGHTS, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to create the user cfs mtab database (%s): %s\n",
				dbName, strerror(errno));
		return (EXIT_FAILURE);
	}

	gdbm_close(dbf);

	return (EXIT_SUCCESS);
}

int mtab_delete(int quiet)
{
	// This function removes the user cfs mtab database
	///////////////////////////////////////////////////

	int rCode;
	char dbName[MAX_DBNAME_SIZE];

	// Get the user cfs mtab database name
	rCode = get_db_name(dbName, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to get the user cfs mtab database name\n");
		return (EXIT_FAILURE);
	}

	// Remove the database  
	rCode = unlink((const char *) dbName);

	if ((rCode == -1) && (errno != ENOENT))
	{
		// It's not an error if the database doesn't exist
		// Here we have an other problem...

		if (!quiet)
			fprintf(stderr,
				"UNABLE to remove the user cfs mtab database (%s): %s\n",
				dbName, strerror(errno));
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}

int create_cfs_fstab(int quiet)
{
	// This function creates the user cfs fstab file.
	// It adds the defined header at the beginning of this file.
	////////////////////////////////////////////////////////////

	int rCode;
	FILE *fd;
	char cfs_fstab_location[CFS_FSTAB_LOCATION_MAX_SIZE];

	// Get the user cfs fstab file name
	rCode = get_file_name(cfs_fstab_location, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to get the user cfs fstab file name\n");
		return (EXIT_FAILURE);
	}

	// Open and create the file. With the w+ option, 
	// the file will be truncated if it already exists.
	fd = fopen(cfs_fstab_location, "w+");

	if (fd == NULL)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to create the user cfs fstab file (%s): %s\n",
				cfs_fstab_location, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Add the defined (in config.h) header
	fprintf(fd, "%s\n", CFS_FSTAB_1_LINE);
	fprintf(fd, "%s\n", CFS_FSTAB_2_LINE);
	fprintf(fd, "%s\n", CFS_FSTAB_3_LINE);
	fprintf(fd, "%s\n", CFS_FSTAB_4_LINE);
	fprintf(fd, "%s\n", CFS_FSTAB_5_LINE);
	fprintf(fd, "%s\n", CFS_FSTAB_6_LINE);

	fclose(fd);

	return (EXIT_SUCCESS);
}

int is_cfs_fstab_exist(int *result, int quiet)
{
	// This function set result to 1 is the user cfs fstab 
	// file exists; set to 0 else.
	//////////////////////////////////////////////////////

	int rCode;
	struct stat finfo;
	char cfs_fstab_location[CFS_FSTAB_LOCATION_MAX_SIZE];

	// Get the user cfs fstab name
	rCode = get_file_name(cfs_fstab_location, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to get the user cfs fstab file name\n");
		return (EXIT_FAILURE);
	}

	// Try to stat this file
	rCode = lstat(cfs_fstab_location, &finfo);

	if (rCode == -1)
	{
		if (errno != ENOENT)
		{
			// It's an abnormal error
			if (!quiet)
				fprintf(stderr,
					"UNABLE to stat the user cfs fstab file (%s): %s\n",
					cfs_fstab_location, strerror(errno));
			return (EXIT_FAILURE);
		}

		else
			*result = 0;		// The file doesn't exist
	}

	else
		*result = 1;			// The file exists

	return (EXIT_SUCCESS);
}

int add_cfs_fstab_entry(char *cipheredDirName, int quiet)
{
	// This function is used to add an new entry in the user cfs 
	// fstab file.
	// This line has defaults values based on cipheredDirName and 
	// explained next.
	/////////////////////////////////////////////////////////////  

	FILE *fd;
	int rCode;
	char *cp;
	char clearDirName[CFS_DIR_NAME_SIZE];
	char absolutePath[CFS_DIR_NAME_SIZE];
	char cfs_fstab_location[CFS_FSTAB_LOCATION_MAX_SIZE];

	// Get the user cfs fstab file name
	rCode = get_file_name(cfs_fstab_location, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to get the user cfs fstab file name\n");
		return (EXIT_FAILURE);
	}

	// Open the user cfs fstab file in append mode
	fd = fopen(cfs_fstab_location, "a");

	if (fd == NULL)
	{
		if (!quiet)
			fprintf(stderr,
				"UNABLE to open the user cfs fstab file (%s): %s\n",
				cfs_fstab_location, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Add the new entry
	// By default: * ciphered directory name is the absolute path
	//               correspondind to the given directory name
	//             * clear name = directory name (without its path)
	//             * no auto attached
	///////////////////////////////////////////////////////////////

	// Get the directory name
	/////////////////////////

	// Search the last '/'
	cp = strrchr(cipheredDirName, '/');

	if (cp == NULL)
	{
		// No '/' found, clear name is just the cipheredDirName 
		strncpy(clearDirName, cipheredDirName, CFS_FSTAB_LOCATION_MAX_SIZE);
		clearDirName[strlen(cipheredDirName)] = '\0';
	}
	else
	{
		// Clear name is the dirctory name, situated after the last '/'
		strncpy(clearDirName, cp + 1, CFS_FSTAB_LOCATION_MAX_SIZE);
		clearDirName[strlen(cp + 1)] = '\0';
	}

	// Get the absolute path
	rCode = get_absolute_path(cipheredDirName, absolutePath, quiet);

	if (rCode == EXIT_FAILURE)
	{
		fclose(fd);
		return (EXIT_FAILURE);
	}

	// Add the new line in the user cfs fstab file
	fprintf(fd, "\n%s\t%s\t%s\n", absolutePath, clearDirName, "noauto");

	fclose(fd);

	return (EXIT_SUCCESS);
}

int give_a_cfs_fstab_entry(char *cipheredDirName, char *clearDirName,
	int lineNr, int *found, int quiet)
{
	// This function stores the content of the lineNr line of the user 
	// cfs fstab file in cipheredDirName and clearDirName.
	// We just consider lines whose auto attach flag is set to auto.
	// The flag found is set to 1 if line is found, and set to 0 else.
	//////////////////////////////////////////////////////////////////

	FILE *fd;
	int rCode;
	char *cp1 = NULL;
	char *cp2 = NULL;
	char *fgetsRCode;
	int currentLineNr = 0;
	int numberOfUselessLines = 0;
	char buffer[2 * CFS_FSTAB_LOCATION_MAX_SIZE];
	char cfs_fstab_location[CFS_FSTAB_LOCATION_MAX_SIZE];

	// Get the user cfs fstab file name
	rCode = get_file_name(cfs_fstab_location, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to get the user cfs fstab file name\n");
		return (EXIT_FAILURE);
	}

	// Open the user cfs fstab file in reading
	fd = fopen(cfs_fstab_location, "r");

	if (fd == NULL)
	{
		if (errno == ENOENT)
		{
			// The user cfs fstab file doesn't exist, it's not a problem
			*found = 0;
			return (EXIT_SUCCESS);
		}
		else
		{
			// But here, it's a problem!!
			if (!quiet)
				fprintf(stderr,
					"UNABLE to open the user cfs fstab file (%s): %s\n",
					cfs_fstab_location, strerror(errno));
			return (EXIT_FAILURE);
		}
	}

	// Select the line number lineNr, it's the lineNr line
	// having the third field set to auto.
	// Commented lines don't enter in count of line.
	// We remenber the number of useless lines skipped to be 
	// able to know the 'absolut' line number.
	////////////////////////////////////////////////////////

	// Get the first line
	fgetsRCode = fgets(buffer, 2 * CFS_FSTAB_LOCATION_MAX_SIZE, fd);

	while (fgetsRCode != NULL)
	{
		// These are commentaries or blank lines
		if ((buffer[0] == '#') || (buffer[0] == '\n'))
			numberOfUselessLines++;

		else
		{
			// Now, parse the line to get its fields
			// Expecting this format:
			// ciphered_directory <tab> clear_name <tab> auto_attach (auto | noauto)
			////////////////////////////////////////////////////////////////////////

			// Search the first tabulation
			cp1 = strchr(buffer, '\t');

			if (cp1 == NULL)
			{
				// No tabulation found, this line has a wrong format
				if (!quiet)
					fprintf(stderr,
						"ERROR: wrong format of line %d in the user cfs fstab file\n",
						currentLineNr + numberOfUselessLines);
				fclose(fd);
				return (EXIT_FAILURE);
			}

			// Copy the ciphered directory in cipheredDirName
			strncpy(cipheredDirName, buffer, cp1 - buffer);
			cipheredDirName[cp1 - buffer] = '\0';

			// Try to get the clear name  by searching the next tabulation
			cp2 = strchr(cp1 + 1, '\t');

			if (cp1 == NULL)
			{
				// No tabulation found, this line has a wrong format
				if (!quiet)
					fprintf(stderr,
						"ERROR: wrong format of line %d in the user cfs fstab file\n",
						currentLineNr + numberOfUselessLines);
				fclose(fd);
				return (EXIT_FAILURE);
			}

			// Copy the clear name in clearDirname
			strncpy(clearDirName, cp1 + 1, cp2 - cp1 - 1);
			clearDirName[cp2 - cp1 - 1] = '\0';

			// Try to get the auto attach flag, 
			// we are normally at the well place 

			if (strncmp(cp2 + 1, "auto", strlen("auto")) == 0)
			{
				// flag is auto, increase the number of line
				currentLineNr++;

				if (currentLineNr == lineNr)	// Are we on the required line?
					goto OUT;
			}

			else if (strncmp(cp2 + 1, "noauto", strlen("noauto")) == 0)
				numberOfUselessLines++;	// It's a useless line

			else
			{
				// This flag is wrong (flag must be auto or noauto)
				if (!quiet)
					fprintf(stderr,
						"ERROR: wrong auto attach flag at the line %d in the user cfs fstab file\n",
						currentLineNr + numberOfUselessLines);
				fclose(fd);
				return (EXIT_FAILURE);
			}
		}

		// Get the next line
		fgetsRCode = fgets(buffer, 2 * CFS_FSTAB_LOCATION_MAX_SIZE, fd);
	}

  OUT:
	// Test if we have really found the good line

	if (currentLineNr != lineNr)
		*found = 0;				// No, we have exited at the end of the file
	else
		*found = 1;				// Yes, required line found

	fclose(fd);

	return (EXIT_SUCCESS);
}

int get_a_cfs_fstab_clear_name(char *cipheredDirName, char *clearDirName,
	int *found, int quiet)
{
	// This function stores in clearDirName the clear name corresponding 
	// to the cipheredDirName in the user cfs fstab file, and sets
	// result to 1.
	// If no name matchs, result is set to 0.
	////////////////////////////////////////////////////////////////////

	FILE *fd;
	int rCode;
	char *cp1 = NULL;
	char *cp2 = NULL;
	char *fgetsRCode;
	int currentLineNr = 1;
	char absolutePath[CFS_DIR_NAME_SIZE];
	char buffer[2 * CFS_FSTAB_LOCATION_MAX_SIZE];
	char cfs_fstab_location[CFS_FSTAB_LOCATION_MAX_SIZE];

	*found = 0;

	// Get the user cfs fstab file name
	rCode = get_file_name(cfs_fstab_location, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to get the user cfs fstab file name\n");
		return (EXIT_FAILURE);
	}

	// Open (in reading) the user cfs fstab file
	fd = fopen(cfs_fstab_location, "r");

	if (fd == NULL)
	{
		if (errno == ENOENT)
		{
			// The user cfs fstab file doesn't exist, it's not a problem
			*found = 0;
			return (EXIT_SUCCESS);
		}
		else
		{
			// But here, it's a problem!!
			if (!quiet)
				fprintf(stderr,
					"UNABLE to open the user cfs fstab file (%s): %s\n",
					cfs_fstab_location, strerror(errno));
			return (EXIT_FAILURE);
		}
	}

	// Get the absolut path of the ciphered directory
	rCode = get_absolute_path(cipheredDirName, absolutePath, quiet);

	if (rCode == EXIT_FAILURE)
	{
		fclose(fd);
		return (EXIT_FAILURE);
	}

	// Get the clear name corresponding
	// to the ciphered directory name given in parameter
	////////////////////////////////////////////////////

	// Get the first line of the file
	fgetsRCode = fgets(buffer, 2 * CFS_FSTAB_LOCATION_MAX_SIZE, fd);

	while (fgetsRCode != NULL)
	{
		// These are commentaries or blank lines
		if ((buffer[0] == '#') || (buffer[0] == '\n'))
			currentLineNr++;

		else
		{
			// Now, parse the line to get its fields
			// Expecting this format:
			// ciphered_directory <tab> clear_name <tab> auto_attach (auto | noauto)
			////////////////////////////////////////////////////////////////////////

			// Search the first tabulation
			cp1 = strchr(buffer, '\t');

			if (cp1 == NULL)
			{
				// No tabulation found, this line has a wrong format
				if (!quiet)
					fprintf(stderr,
						"ERROR: wrong format of line %d in the user cfs fstab file\n",
						currentLineNr);
				fclose(fd);
				return (EXIT_FAILURE);
			}

			if (strncmp(absolutePath, buffer, cp1 - buffer - 1) == 0)
			{
				// It's the ciphered dir we are loocking for
				////////////////////////////////////////////

				// Try to get the clear name by searching the next tabulation
				cp2 = strchr(cp1 + 1, '\t');

				if (cp1 == NULL)
				{
					// No tabulation found, this line has a wrong format
					if (!quiet)
						fprintf(stderr,
							"ERROR: wrong format of line %d in the user cfs fstab file\n",
							currentLineNr);
					fclose(fd);
					return (EXIT_FAILURE);
				}

				// Copy the clear name in clearDirname
				strncpy(clearDirName, cp1 + 1, cp2 - cp1 - 1);
				clearDirName[cp2 - cp1 - 1] = '\0';

				*found = 1;
				fclose(fd);

				return (EXIT_SUCCESS);
			}
		}
		// Get the next line
		fgetsRCode = fgets(buffer, 2 * CFS_FSTAB_LOCATION_MAX_SIZE, fd);
	}

	fclose(fd);

	return (EXIT_SUCCESS);
}

int get_db_name(char *dbName, int quiet)
{
	// This function stores in dbName the user cfs mtab name
	////////////////////////////////////////////////////////

	struct passwd *structPasswd;

	// Get the userName by his uid in the passwd file
	structPasswd = getpwuid(getuid());

	if (structPasswd == NULL)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to find the user in the passwd file: %s",
				strerror(errno));
		return (EXIT_FAILURE);
	}

	// Construct the database name: /tmp/cfs_SC_mtab_username
	sprintf(dbName, "%s_%s", CFS_MTAB_NAME, structPasswd->pw_name);

	return (EXIT_SUCCESS);
}

int get_file_name(char *fileName, int quiet)
{
	// This function stores in fileName the user cfs fstab name
	///////////////////////////////////////////////////////////

	struct passwd *structPasswd;

	// Get the user home directory by his uid in the passwd file
	structPasswd = getpwuid(getuid());

	if (structPasswd == NULL)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to find the user in the passwd file: %s",
				strerror(errno));
		return (EXIT_FAILURE);
	}

	// Construct the file name: /user_home_directory/cfs_fstab
	sprintf(fileName, "%s/%s", structPasswd->pw_dir, CFS_FSTAB_NAME);

	return (EXIT_SUCCESS);
}

int get_absolute_path(char *relativePath, char *absolutePath, int quiet)
{
	// This function converts the ciphered directory relative path given by the
	// user to an absolute path. This is done just using the library function
	// realpath(...).
	///////////////////////////////////////////////////////////////////////////

	char *rCode;

	rCode = realpath(relativePath, absolutePath);

	if (rCode == NULL)
	{
		// Problem during the conversion
		if (!quiet)
		{
			fprintf(stderr,
				"UNABLE to convert the ciphered directory relative path (%s)\n",
				relativePath);
			fprintf(stderr, "in an absolute path: %s\n", strerror(errno));
		}

		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
