/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: pid_list.c,v 1.2 2002/03/07 10:33:05 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <signal.h>

#include "config.h"
#include "global.h"
#include "pam_smartcard_utils.h"
#include "pid_list.h"

// The next function are used to manage the autolock pid file.
// This file is used to store the pid of the autolock daemon
// and the list of the pid of the currently opened sessions. 
// One pid per line.
//////////////////////////////////////////////////////////////

int listInit(char *userName)
{
	// This function is used to create and initialize the 
	// autolock pid file.
	//////////////////////////////////////////////////////

	int rCode;
	FILE *fd;
	struct passwd *structPasswd;

	// Open the file
	fd = fopen(PID_AUTOLOCK_FILE, "w");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Change the autolock pid file owner to the user
	/////////////////////////////////////////////////

	// Get the user uid using the passwd file
	structPasswd = getpwnam(userName);

	if (structPasswd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to find the user in the passwd file: %s",
			strerror(errno));
		return (PAM_SESSION_ERR);
	}

	// Change the autolock pid file owner
	rCode = chown(PID_AUTOLOCK_FILE, structPasswd->pw_uid, -1);

	if (rCode == -1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to change the owner of the autolock pid file: %sn",
			strerror(errno));
		return (PAM_SESSION_ERR);
	}

	fclose(fd);
	return (EXIT_SUCCESS);
}

int setAutolockPid(int autolockPid)
{
	// This function writes the autolock pid at the 
	// beginning of the autolock pid file.
	//////////////////////////////////////////////////////

	int nbWrote;
	FILE *fd;

	// Open the file
	fd = fopen(PID_AUTOLOCK_FILE, "w+");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Write the autolock pid
	nbWrote = fprintf(fd, "%d\n", autolockPid);

	if (nbWrote < 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to write the autolock pid in autolock pid file (%s)",
			PID_AUTOLOCK_FILE);
		return (EXIT_FAILURE);
	}

	fclose(fd);
	return (EXIT_SUCCESS);
}

int getAutolockPid(int *pid)
{
	// The function retrieves the autolock pid int the autolock pid
	// file and stores it in pid.
	///////////////////////////////////////////////////////////////

	FILE *fd;
	int rCode;
	struct stat statbuf;

	// Open the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "r");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// If the size of the autolock pid file is 0, the autolock
	// doesn't run.
	rCode = stat(PID_AUTOLOCK_FILE, &statbuf);

	if (rCode == -1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to stat the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		fclose(fd);
		return (EXIT_FAILURE);
	}

	else if (statbuf.st_size == 0)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"ERROR: the autolock daemon seems not to be running\n");
		fclose(fd);
		return (EXIT_FAILURE);
	}

	// The autolock pid is the first line of the file
	rCode = fscanf(fd, "%d", pid);

	if (rCode != 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to read the autolock pid in the autolock pid file (%s)",
			PID_AUTOLOCK_FILE);
		fclose(fd);
		return (EXIT_FAILURE);
	}

	fclose(fd);
	return (EXIT_SUCCESS);
}

int getSessionPid(int *pid, int *result, int nr)
{
	// This function retyurn in pid the nr opened session pid
	// registred in the autolock pid file.
	// Result is set to 1 if it is found, set to 0 else
	/////////////////////////////////////////////////////////

	int rCode;
	int i = 0;
	FILE *fd;
	int pidReaded;

	*result = 0;

	// Open the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "r");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Skip the autolock pid
	rCode = fscanf(fd, "%d", &pidReaded);

	if (rCode != 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to skip the autolock pid in the autolock pid file (%s)",
			PID_AUTOLOCK_FILE);
		fclose(fd);
		return (EXIT_FAILURE);
	}

	// Read the first opened session pid
	rCode = fscanf(fd, "%d", &pidReaded);

	while (rCode == 1)
	{
		if (i++ == nr)
		{
			// nr pid found in the list
			*result = 1;
			*pid = pidReaded;
		}

		// Read the next opened session pid
		rCode = fscanf(fd, "%d", &pidReaded);
	}

	fclose(fd);
	return (EXIT_SUCCESS);
}

int listEmpty(int *result)
{
	// This function set result to 1 if the opened session list of 
	// the autolock pid is empty; set result to 0 else.
	//////////////////////////////////////////////////////////////

	int rCode;
	FILE *fd;
	int pidReaded;

	*result = 0;

	// Open the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "r");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Skip the autolock pid
	rCode = fscanf(fd, "%d", &pidReaded);

	if (rCode != 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to skip the autolock pid in the autolock pid file (%s)",
			PID_AUTOLOCK_FILE);
		fclose(fd);
		return (EXIT_FAILURE);
	}

	// Try to read a opened session pid
	rCode = fscanf(fd, "%d", &pidReaded);

	if (rCode != 1)
	{
		// The list is empty
		*result = 1;

		fclose(fd);
		return (EXIT_SUCCESS);
	}

	// The list is not empty  
	fclose(fd);
	return (EXIT_SUCCESS);
}

int putPidInList(int pid)
{
	// This function put pid at the end of the autolock pid file
	////////////////////////////////////////////////////////////

	FILE *fd;
	int nbWrote;

	// Open the autolock pid file an the appen mode
	fd = fopen(PID_AUTOLOCK_FILE, "a+");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Write the pid
	nbWrote = fprintf(fd, "%d\n", pid);

	if (nbWrote < 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to write in the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		fclose(fd);
		return (EXIT_FAILURE);
	}

	fclose(fd);
	return (EXIT_SUCCESS);
}

int deletePidInList(int pid)
{
	// This function delete pid in the autolock pid file.
	// If Pid not in the list nothing is done.
	////////////////////////////////////////////////////

	int rCode;
	FILE *fd;
	char tmpTab[BUF_TMP_SIZE];
	int pidReaded;
	int posInTmpTab = 0;

	memset(tmpTab, '\0', BUF_TMP_SIZE);

	// Open the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "r");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Skip the autolock pid (it's the first bytes)
	// to be positionated at the begining of the session pid list.
	// Copy it in the temporary tab
	rCode = fscanf(fd, "%d", &pidReaded);

	if (rCode != 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to skip the autolock pid in the autolock pid file (%s)",
			PID_AUTOLOCK_FILE);
		fclose(fd);
		return (EXIT_FAILURE);
	}

	posInTmpTab += sprintf(tmpTab + posInTmpTab, "%d\n", pidReaded);

	// Read the first opened session pid
	rCode = fscanf(fd, "%d", &pidReaded);

	while (rCode == 1)
	{
		if (pidReaded == pid)
		{
			// Pid found in the list
			// Don't write the deleted pid in the temporary tab
		}
		else
		{
			// Write the pid in the temporary tab
			posInTmpTab += sprintf(tmpTab + posInTmpTab, "%d\n", pidReaded);
		}

		// Read the next opened session pid
		rCode = fscanf(fd, "%d", &pidReaded);
	}

	fclose(fd);

	// Reopen the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "w");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Copy the temporary tab in the new autolock pid file
	fprintf(fd, "%s", tmpTab);

	fclose(fd);

	return (EXIT_SUCCESS);
}

int pidInList(int pid, int *result)
{
	// This function set result to 1 if pid is registred
	// in the autolock pid file; set result to 0 else.
	////////////////////////////////////////////////////

	int rCode;
	FILE *fd;
	int pidReaded;

	*result = 0;

	// Open the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "r");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Skip the autolock pid (it's the first bytes)
	// to be positionated at the begining of the session pid list.
	rCode = fscanf(fd, "%d", &pidReaded);

	if (rCode != 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to skip the autolock pid in the autolock pid file (%s)",
			PID_AUTOLOCK_FILE, strerror(errno));
		fclose(fd);
		return (EXIT_FAILURE);
	}

	// Read the first opened session pid
	rCode = fscanf(fd, "%d", &pidReaded);

	while (rCode == 1)
	{
		if (pidReaded == pid)
		{
			// Pid found in the list
			*result = 1;
			fclose(fd);
			return (EXIT_SUCCESS);
		}

		// Read the next opened session pid
		rCode = fscanf(fd, "%d", &pidReaded);
	}

	// Pid not found in the list
	fclose(fd);
	return (EXIT_SUCCESS);
}

int cleanupList(void)
{
	// This function check the list in order to verify that all the 
	// registred sessions are still alive
	///////////////////////////////////////////////////////////////

	int pid_nbr;
	int pid;
	int rCode;
	int result;

	// Get the first opened session pid
	pid_nbr = 0;
	rCode = getSessionPid(&pid, &result, pid_nbr);

	if (rCode == EXIT_FAILURE)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR,
			"UNABLE to get a session pid from the autolock pid file");
		return (EXIT_FAILURE);
	}

	while (result == 1)
	{
		// To test if the session number pid is still alive,
		// We try to send it the null signal 
		rCode = kill(pid, 0);

		if ((rCode == -1) && (errno == ESRCH))
		{
			// This session doesn't exist
			// Remove it from the list

			pam_log(AUTOLOCK_IDENT, LOG_ERR,
				"Cleaning pid %d in the autolock pid file", pid);

			if (deletePidInList(pid) == EXIT_FAILURE)
			{
				pam_log(AUTOLOCK_IDENT, LOG_ERR,
					"Unable to delete a session in the autolock pid file");
				exit(EXIT_FAILURE);
			}

			// we removed one pid in the list
			// the pid number of the "next" one does not change
		}
		else
			// Read the next pid in the list
			pid_nbr++;

		// Get the next opened session pid
		rCode = getSessionPid(&pid, &result, pid_nbr);

		if (rCode == EXIT_FAILURE)
		{
			pam_log(AUTOLOCK_IDENT, LOG_ERR,
				"UNABLE to get an opened session pid from the autolock pid file");
			return (EXIT_FAILURE);
		}
	}

	return (EXIT_SUCCESS);
}
