/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: pam_smartcard_utils.c,v 1.9 2002/03/12 10:33:04 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */
#include <stdio.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <gdbm.h>
#include <gmp.h>
#include <pwd.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <rpcsvc/ypclnt.h>
#include <rpcsvc/yp_prot.h>
#include <linux/soundcard.h>

#include "config.h"
#include "global.h"
#include "debug_config.h"
#include "debug.h"

#include "pam_smartcard_utils.h"
#include "crypto_functions.h"
#include "publickey_db_functions.h"

// Internal functions
/////////////////////

// Verify that the public key have the rigth administrator signature
int verify_public_key_signature(char *publicKey, char *signature,
	char *userName);

// This function hashs (using MD5) and padd (using PKCS1 padding) the data
void PKCS_1_hash_padding(unsigned char *data, int dataSize,
	unsigned char *result);

int nis_get_user_info(char *userName, mpz_t *pubExpo, mpz_t *pubModulo)
{
	// This function retreives the user public key from the NIS map and 
	// verify it's signature
	// This function does NOT malloc any memory and return the public key
	// components.
	////////////////////////////////////////////////////////////////////////

	int rCode = 0, dataLen;
	char *nisDomain, *data;
	char publicKey[RSA_KEY_SIZE_IN_TXT + 1];
	char signature[CERT_RSA_KEY_SIZE_IN_TXT];

	// Get the user public key from NIS
	///////////////////////////////////

	// Get the domain name
	rCode = yp_get_default_domain(&nisDomain);

	if (rCode != 0)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"Get NIS default domain error: %s", yperr_string(rCode));
		return (PAM_AUTHINFO_UNAVAIL);
	}

	// Search for the user in the map
	rCode = yp_match(nisDomain, NIS_KEY_MAP_NAME, userName,
		strlen(userName), &data, &dataLen);
	if (rCode != 0)
	{
		if (rCode == 4)
		{
			// User not found
			pam_log(PAM_SMARTCARD, LOG_ALERT,
				"User %s not found in the NIS map %s\n", userName);
			return (PAM_AUTHINFO_UNAVAIL);
		}
		else
		{
			// Other error
			pam_log(PAM_SMARTCARD, LOG_ALERT,
				"NIS match user error: %s", yperr_string(rCode));
			return (PAM_AUTHINFO_UNAVAIL);
		}
	}

	// Verify the data size
	// Size is RSA_KEY_SIZE_IN_TXT + size of the signature : CERT_RSA_KEY_SIZE_IN_TXT 

	if (dataLen != RSA_KEY_SIZE_IN_TXT + CERT_RSA_KEY_SIZE_IN_TXT)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"Problem with NIS: Size of public key readed differs from the correct size");
		return (PAM_AUTHINFO_UNAVAIL);
	}

	// Verify the public key signature
	//////////////////////////////////

	// Separate the signature from the key

	memcpy(publicKey, data, RSA_KEY_SIZE_IN_TXT);
	memcpy(signature, data + RSA_KEY_SIZE_IN_TXT, CERT_RSA_KEY_SIZE_IN_TXT);
	publicKey[RSA_KEY_SIZE_IN_TXT] = '\0';


	// Verify the signature
	rCode = verify_public_key_signature(publicKey, signature, userName);
	if (rCode == EXIT_FAILURE)
		return (PAM_AUTHINFO_UNAVAIL);

	// Convert the public modulo and the public exposant in a mpz number

	mpz_init(*pubExpo);
	mpz_init(*pubModulo);

	mpz_init_set_str(*pubExpo, RSA_PUBLIC_EXPOSANT, 16);
	mpz_init_set_str(*pubModulo, publicKey, 16);

	return (PAM_SUCCESS);
}

int localdb_get_user_info(char *userName, char *localDbName,
	mpz_t *pubExpo, mpz_t *pubModulo)
{
	// This function retreives the user public key from the local
	// database and verify it's signature
	// This function does NOT malloc any memory and return the public key
	// components.
	////////////////////////////////////////////////////////////////////////

	int i, rCode;
	GDBM_FILE dbf;
	datum dbKey, data;
	struct stat finfo;
	char dataTxt[CERT_RSA_KEY_SIZE_IN_TXT + RSA_KEY_SIZE_IN_TXT + 1];
	char publicKey[RSA_KEY_SIZE_IN_TXT + 1];
	char signature[CERT_RSA_KEY_SIZE_IN_TXT +1];

	// Get the user public key from specified local database
	////////////////////////////////////////////////////////

	// First check : Does the file exist ?? 
	rCode = lstat(localDbName, &finfo);

	if (rCode == -1)
	{
		// Can't stat the file... 
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"Couldn't stat database file %s", localDbName);
		return (PAM_SERVICE_ERR);
	}

	// Second check : Is it a regular file ?? 
	if (!S_ISREG(finfo.st_mode))
	{
		// File is not a regular file 
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"Database file %s is not a regular file", localDbName);
		return (PAM_SERVICE_ERR);
	}

	// Third check : Check the permissions
	if (finfo.st_mode != 0100744)
	{
		// Wrong permissions
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"Wrong file permissions for database file %s", localDbName);
		return (PAM_SERVICE_ERR);
	}

	// Open the dbm database file
	dbf = gdbm_open(localDbName, DB_BLOC_SIZE, GDBM_READER,
		OPEN_CREATE_DB_RIGHTS, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		// Unable to open the file
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"Unable to open the public key database %s: %s\n",
			localDbName, strerror(errno));
		return (PAM_AUTHINFO_UNAVAIL);
	}

	// Store the user name in a valid data for dbm file
	dbKey.dptr = userName;
	dbKey.dsize = strlen(userName);

	// Retrieve the public key which correspond to the user
	data = gdbm_fetch(dbf, dbKey);

	if (data.dptr == NULL)
	{
		// User don't exists in the database
		pam_log(PAM_SMARTCARD, LOG_INFO,
			"User %s not found in the public key database %s\n", userName,
			localDbName);

		gdbm_close(dbf);
		return (PAM_AUTHINFO_UNAVAIL);
	}

	// Verify the data size
	// Size is RSA_KEY_SIZE_IN_BYTES + size of the signature (CERT_RSA_KEY_SIZE_IN_BYTES) 

	if (data.dsize != (RSA_KEY_SIZE_IN_BYTES + CERT_RSA_KEY_SIZE_IN_BYTES))
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"Problem with public key database: size of public key readed differs from the correct size");

		gdbm_close(dbf);
		return (PAM_AUTHINFO_UNAVAIL);
	}

	// Close the dbm file
	gdbm_close(dbf);

	// Verify the public key signature
	//////////////////////////////////

	for (i = 0; i < RSA_KEY_SIZE_IN_BYTES + CERT_RSA_KEY_SIZE_IN_BYTES; i++)
		sprintf(dataTxt + 2 * i, "%02x", (unsigned char) ((data.dptr)[i]));

	memcpy(publicKey, dataTxt, RSA_KEY_SIZE_IN_TXT);
	publicKey[RSA_KEY_SIZE_IN_TXT] = '\0';

	memcpy(signature, dataTxt + RSA_KEY_SIZE_IN_TXT, CERT_RSA_KEY_SIZE_IN_TXT);
	signature[CERT_RSA_KEY_SIZE_IN_TXT] = '\0';

	rCode = verify_public_key_signature(publicKey, signature, userName);
	if (rCode == EXIT_FAILURE)
		return (PAM_AUTHINFO_UNAVAIL);

	// Convert the public modulo and the public exposant in a mpz number

	mpz_init(*pubExpo);
	mpz_init(*pubModulo);

	mpz_init_set_str(*pubExpo, RSA_PUBLIC_EXPOSANT, 16);
	mpz_init_set_str(*pubModulo, publicKey, 16);

	return (PAM_SUCCESS);
}

int pam_parse(int argc, const char *argv[], int *ctrl, char *localDbName)
{
	// This function parses the argument optionnaly given to this module in 
	// the pam configuration file.
	// This function is designed to recognize numerous options but
	// actually, just the local option is used...
	// Recognized options are available in ctrl.
	// If a specific local database is given, its name is stored in
	// localDbName.
	////////////////////////////////////////////////////////////////////////

	*ctrl = 0;

	// By default, we use the default local database
	strncpy(localDbName, DEFAULT_DB_NAME, MAX_LEN_OF_DB_NAME);

	// Step through arguments 
	for (; argc-- > 0; argv++)
	{
		// Generic argument 

		if (!strncmp(*argv, "verbose", strlen("verbose")))
			*ctrl |= PAM_VERBOSE_ARG;
		else if (!strncmp(*argv, "no_warn", strlen("no_warn")))
			*ctrl |= PAM_NWARN_ARG;
		else if (!strncmp(*argv, "use_first_pass", strlen("use_first_pass")))
			*ctrl |= PAM_UFIRST_ARG;
		else if (!strncmp(*argv, "try_first_pass", strlen("try_first_pass")))
			*ctrl |= PAM_TFIRST_ARG;
		else if (!strncmp(*argv, "use_mapped_pass",
				strlen("use_mapped_pass")))
			*ctrl |= PAM_MAPPED_ARG;
		else
		{
			if (!strncmp(*argv, "nis", strlen("nis")))
			{
				// Publics key will be retrivied using NIS
				*ctrl |= PAM_NIS;
			}
			else
			{
				// The argument is consider as a specific local database name
				strncpy(localDbName, *argv, MAX_LEN_OF_DB_NAME);
			}
		}
	}

	return (EXIT_SUCCESS);
}

void pam_log(int processIdent, int err, const char *format, ...)
{
	// This function is used to log in the system /var/log/ file
	// different events.
	// According to processIdent a specific header is inserted
	// before format.
	// err is the level of the error (described in syslog(3)).
	////////////////////////////////////////////////////////////

	va_list args;
	char *ident = NULL;

	// Find the required header
	switch (processIdent)
	{
	case AUTOLOCK_IDENT:
		ident = "autolock ";
		break;

	case PAM_SMARTCARD_AUTH:
		ident = "PAM_smartcard (auth part) ";
		break;

	case PAM_SMARTCARD_SESSION:
		ident = "PAM_smartcard (session part) ";
		break;

	case PAM_SMARTCARD:
		ident = "PAM_smartcard ";
		break;

	case PAM_CFS_SC:
		ident = "PAM_cfs_SC ";
		break;
	}

	// Initialize the argument list
	va_start(args, format);

	// All the events are logged in LOG_AUTH (auth.log file)
	openlog(ident, LOG_CONS | LOG_PID, LOG_AUTH);
	vsyslog(err, format, args);
	closelog();

	// Close the argument list
	va_end(args);
}

int pam_conversate(pam_handle_t * pamh, char *pin)
{
	// This function manages the communication with the user.
	// It uses the PAM conversation mechanism.
	// First it tells to insert his smartcard.
	// And she ask to he for the PIN code (whitch it's stored in pin).
	//////////////////////////////////////////////////////////////////

	int rCode;
	struct pam_conv *conv;
	struct pam_response *resp;
	struct pam_message *mesg[1], msg[1];

	// Get the conversation function stored in the pam handle structure
	rCode = pam_get_item(pamh, PAM_CONV, (const void **) &conv);

	if (rCode != PAM_SUCCESS)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to obtain the conversation function: %s\n",
			pam_strerror(pamh, rCode));
		return (PAM_CONV_ERR);
	}

	// Informative message: Insert the card
	///////////////////////////////////////

	// Fill up the message structure  
	mesg[0] = msg;
	msg[0].msg_style = PAM_TEXT_INFO;	// We give some info, no response required 
	msg[0].msg = INSERT_CARD_MESSAGE;

	// Call the conversation function
	rCode = conv->conv(1, (const struct pam_message **) mesg,
		(struct pam_response **) &resp, conv->appdata_ptr);

	if (rCode != PAM_SUCCESS)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"Conversation FAILURE: %s\n", pam_strerror(pamh, rCode));
		return (PAM_CONV_ERR);
	}

	// OK, now ask for the pin
	//////////////////////////

	// Fill up the message structure
	msg[0].msg_style = PAM_PROMPT_ECHO_OFF;
	msg[0].msg = ENTER_PIN_MESSAGE;

	// Call the conversation function
	rCode = conv->conv(1, (const struct pam_message **) mesg,
		(struct pam_response **) &resp, conv->appdata_ptr);

	if (rCode != PAM_SUCCESS)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"Conversation FAILURE: %s\n", pam_strerror(pamh, rCode));
		return (PAM_CONV_ERR);
	}

	// Store the given PIN code
	strncpy(pin, resp[0].resp, USR_PIN_SIZE);
	pin[USR_PIN_SIZE] = '\0';

	return (PAM_SUCCESS);
}

int generaterandom(unsigned char *bufRandom, unsigned long lengthRequired)
{
	// This function fills up bufRandom with lengthRequired random numbers.
	// This numbers are get using the kernel random number generator
	// /dev/urandom which seems to produce number having good cryptographic
	// properties.
	///////////////////////////////////////////////////////////////////////

	int i, fd, rCode;

	// Open the kernel random number generator
	fd = open("/dev/urandom", O_RDONLY);

	if (fd == -1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the kernel random number generator: %s",
			strerror(errno));
		return (EXIT_FAILURE);
	}

	// Read bytes to fill up the bufRandom
	for (i = 0; i < lengthRequired; i++)
	{
		rCode = read(fd, bufRandom + i, 1);

		if (rCode == -1)
		{
			pam_log(PAM_SMARTCARD, LOG_ERR,
				"UNABLE to read numbers from the kernel random number generator: %s",
				strerror(errno));
			close(fd);
			return (EXIT_FAILURE);
		}
	}
	close(fd);
	return (EXIT_SUCCESS);
}

int is_user_excluded(char *username)
{
	// Parse the autolock daemon excluded users file
	// Return EXIT_SUCCESS is the user is in this file
	// and EXIT_FAILURE else.
	//////////////////////////////////////////////////

	FILE *fd;
	int rCode = 0;
	struct stat finfo;
	char buffer[MAX_LINE_LENGTH];

	// First check : Does the file exist ?? 
	if (lstat(AUTOLOCK_EXCLUDED_FILE_NAME, &finfo))
	{
		// Doesn't matter... Exit
		return (EXIT_FAILURE);
	}

	// Open the excluded users file
	fd = fopen(AUTOLOCK_EXCLUDED_FILE_NAME, "r");

	if (fd == NULL)
	{
		// We can't open the file... Log the problem
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"Can't open file the autolock excluded users file %s: %s\n",
			AUTOLOCK_EXCLUDED_FILE_NAME, strerror(errno));
		return (EXIT_FAILURE);
	}

	// File is open, all is OK.. Look for the user

	while (!feof(fd))
	{
		fgets(buffer, MAX_LINE_LENGTH, fd);

		if (buffer == NULL)
		{
			pam_log(PAM_SMARTCARD, LOG_ERR,
				"Can't read in the autolock excluded users file %s: %s\n",
				AUTOLOCK_EXCLUDED_FILE_NAME, strerror(errno));
			return (EXIT_FAILURE);
		}

		// User matches ?? 
		rCode = strncmp(buffer, username, strlen(username));

		if (rCode == 0)
			break;				// User is found
	}

	fclose(fd);

	// Check if we really have a match
	if (rCode != 0)
		return (EXIT_FAILURE);	// It wasn't a match
	else
		return (EXIT_SUCCESS);
}

int verify_public_key_signature(char *publicKey, char *signature,
	char *userName)
{
	int i, rCode;
	int ret;
	FILE *fd;
	mpz_t certPubExpo, certPubModulo;
	unsigned char signatureInBytes[CERT_RSA_KEY_SIZE_IN_BYTES];
	char certKeyTxt[CERT_RSA_KEY_SIZE_IN_TXT + 1];
	char data[RSA_KEY_SIZE_IN_TXT + USER_NAME_MAX_SIZE];

	// Get the administrator certification public key
	/////////////////////////////////////////////////

	// Open the root certification key file
	fd = fopen(ROOT_CERT_FILE_NAME, "r");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ALERT,
			"Unable to open the administrator certification public key file (%s): %s\n",
			ROOT_CERT_FILE_NAME, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Read the root public key
	for (i = 0; i < CERT_RSA_KEY_SIZE_IN_TXT; i++)
	{
		rCode = fscanf(fd, "%c", certKeyTxt + i);

		if (rCode != 1)
		{
			pam_log(PAM_SMARTCARD, LOG_ALERT,
				"Unable to read the root certification public key: %s\n",
				strerror(errno));
			return (EXIT_FAILURE);
		}
	}

	fclose(fd);

	// Convert the public modulo and the public exposant in a mpz number
	////////////////////////////////////////////////////////////////////

	certKeyTxt[CERT_RSA_KEY_SIZE_IN_TXT] = '\0';

	mpz_init(certPubExpo);
	mpz_init(certPubModulo);

	mpz_init_set_str(certPubExpo, RSA_PUBLIC_EXPOSANT, 16);
	mpz_init_set_str(certPubModulo, certKeyTxt, 16);

	// Convert the signature in a byte form
	///////////////////////////////////////

	for (i = 0; i < CERT_RSA_KEY_SIZE_IN_BYTES; i++)
	{
		int tmp;
		sscanf(signature + 2 * i, "%02x", &tmp);
		signatureInBytes[i] = (unsigned char) tmp;
	}
	DEBUG_INFO2("signature: %s", signature);
	DEBUG_INFO2("public key: %s", publicKey);
	DEBUG_INFO2("user name: %s", userName);

	// Data signed are concatenation of the username and the user public key
	////////////////////////////////////////////////////////////////////////

	memset(data, 0, USER_NAME_MAX_SIZE);
	strncpy(data, userName, USER_NAME_MAX_SIZE);
	memcpy(data + USER_NAME_MAX_SIZE, publicKey, RSA_KEY_SIZE_IN_TXT);

	// Now, we have the elements, can call the function verify_signature 
	////////////////////////////////////////////////////////////////////  

	if (verify_signature(&certPubExpo, &certPubModulo, signatureInBytes,
			data, USER_NAME_MAX_SIZE + RSA_KEY_SIZE_IN_TXT) == EXIT_SUCCESS)
		// Signature is OK
		ret = EXIT_SUCCESS;
	else
	{
		// Signature is not OK

		pam_log(PAM_SMARTCARD, LOG_ERR,
			"Wrong signature of the user public key");
		ret = EXIT_FAILURE;
	}

	mpz_clear(certPubExpo);
	mpz_clear(certPubModulo);

	return ret;
}

void PKCS_1_hash_padding(unsigned char *data, int dataSize,
	unsigned char *result)
{
	// This function hashs data using MD5 and Padd the digest using
	// PKCS1. The resul is stored in...result!
	///////////////////////////////////////////////////////////////

	int i;
	struct MD5Context md5Context;
	unsigned char digest[DIGEST_SIZE];

	// Hash the data using the MD5 algorithm
	////////////////////////////////////////

	MD5Init(&md5Context);
	MD5Update(&md5Context, data, dataSize);
	MD5Final(digest, &md5Context);

	// Padd the hashed data with the PKCS1 padding
	//////////////////////////////////////////////

	// Specific PKCS1 header
	result[0] = 0x00;
	result[1] = 0x01;

	for (i = 2; i < 29 + HASH_PADDING_SIZE - 64; i++)
		result[i] = 0xFF;

	result[29 + HASH_PADDING_SIZE - 64] = 0x00;
	result[30 + HASH_PADDING_SIZE - 64] = 0x30;
	result[31 + HASH_PADDING_SIZE - 64] = 0x20;
	result[32 + HASH_PADDING_SIZE - 64] = 0x30;
	result[33 + HASH_PADDING_SIZE - 64] = 0x0C;
	result[34 + HASH_PADDING_SIZE - 64] = 0x06;
	result[35 + HASH_PADDING_SIZE - 64] = 0x08;
	result[36 + HASH_PADDING_SIZE - 64] = 0x2A;
	result[37 + HASH_PADDING_SIZE - 64] = 0x86;
	result[38 + HASH_PADDING_SIZE - 64] = 0x48;
	result[39 + HASH_PADDING_SIZE - 64] = 0x86;
	result[40 + HASH_PADDING_SIZE - 64] = 0xF7;
	result[41 + HASH_PADDING_SIZE - 64] = 0x0D;
	result[42 + HASH_PADDING_SIZE - 64] = 0x02;
	result[43 + HASH_PADDING_SIZE - 64] = 0x05;
	result[44 + HASH_PADDING_SIZE - 64] = 0x05;
	result[45 + HASH_PADDING_SIZE - 64] = 0x00;
	result[46 + HASH_PADDING_SIZE - 64] = 0x04;
	result[47 + HASH_PADDING_SIZE - 64] = 0x10;

	// Hashed data at the end
	strncpy((char *) (result + 48 + HASH_PADDING_SIZE - 64), (char *) digest,
		16);
}

int verify_signature(mpz_t *pubExpo, mpz_t *pubModulo, unsigned char *signature,
	unsigned char *initialData, int initialDataSize)
{
	int i;
	int ret;
	mpz_t mpzCalculated;
	mpz_t mpzExpected;
	mpz_t mpzSignature;
	unsigned char bufTmp[HASH_PADDING_SIZE_IN_TXT + 5];
	unsigned char expected[HASH_PADDING_SIZE];

	// Convert the signature into a mpz number
	//////////////////////////////////////////

	// Initialize the big num
	mpz_init(mpzSignature);
	mpz_init(mpzCalculated);
	mpz_init(mpzExpected);

	// Litle bug in libgmp: buffer overflow
	// So, we use a buffer bigger in which the 4 first bytes are '0'

#if 0
	bufTmp[0] = bufTmp[1] = bufTmp[2] = bufTmp[3] = '0';
	for (i = 0; i < HASH_PADDING_SIZE; i++)
		sprintf(bufTmp + 4 + 2 * i, "%02x", signature[i]);
	bufTmp[HASH_PADDING_SIZE_IN_TXT + 4] = '\0';
#endif
	for (i = 0; i < HASH_PADDING_SIZE; i++)
		sprintf(bufTmp + 2 * i, "%02x", signature[i]);
	bufTmp[HASH_PADDING_SIZE_IN_TXT] = '\0';

	mpz_set_str(mpzSignature, bufTmp, 16);

	// Decrypt the signed number with the public key 
	// Just here is the operation which perform the RSA deciphering
	///////////////////////////////////////////////////////////////

	mpz_powm(mpzCalculated, mpzSignature, *pubExpo, *pubModulo);

	// Hash and padd the initial data
	/////////////////////////////////

	DEBUG_INFO2("Initial data size: %d", initialDataSize);
	DEBUG_XXD("initial data: ", initialData, initialDataSize);

	PKCS_1_hash_padding(initialData, initialDataSize, expected);

	DEBUG_XXD("expected: ", expected, sizeof(expected));

	// Convert the result in an mpz number
	/////////////////////////////////////

	// Always the same bug in libgmp : buffer overflow
	// So, we use a buffer bigger in which the 4 prime bytes are '0'

#if 0
	bufTmp[0] = bufTmp[1] = bufTmp[2] = bufTmp[3] = '0';
	for (i = 0; i < HASH_PADDING_SIZE; i++)
		sprintf(bufTmp + 4 + 2 * i, "%02x", expected[i]);
	bufTmp[HASH_PADDING_SIZE_IN_TXT + 4] = '\0';
#endif
	for (i = 0; i < HASH_PADDING_SIZE; i++)
		sprintf(bufTmp + 2 * i, "%02x", expected[i]);
	bufTmp[HASH_PADDING_SIZE_IN_TXT] = '\0';

	mpz_set_str(mpzExpected, bufTmp, 16);

	DEBUG_INFO2("pubExpo: %s", mpz_get_str(NULL, 16, *pubExpo));
	DEBUG_INFO2("pubModulo: %s", mpz_get_str(NULL, 16, *pubModulo));
	DEBUG_INFO2("mpzSignature: %s", mpz_get_str(NULL, 16, mpzSignature));
	DEBUG_INFO2("mpzCalculated: %s", mpz_get_str(NULL, 16, mpzCalculated));
	DEBUG_INFO2("mpzExpected: %s", mpz_get_str(NULL, 16, mpzExpected));

	// Test if the hashed random number and the hashed card 
	// cipher/decipher random number are the same
	///////////////////////////////////////////////////////

	if (mpz_cmp(mpzCalculated, mpzExpected) == 0)
		ret = EXIT_SUCCESS;	// Good signature
	else
		ret = EXIT_FAILURE;	// Wrong signature

	mpz_clear(mpzSignature);
	mpz_clear(mpzCalculated);
	mpz_clear(mpzExpected);

	return ret;
}

// The next function are used to manage the autolock pid file.
// This file is used to store the pid of the autolock daemon
// and the list of the pid of the currently opened sessions. 
// One pid per line.
//////////////////////////////////////////////////////////////

int listInit(char *userName)
{
	// This function is used to create and initialize the 
	// autolock pid file.
	//////////////////////////////////////////////////////

	int rCode;
	FILE *fd;
	struct passwd *structPasswd;

	// Open the file
	fd = fopen(PID_AUTOLOCK_FILE, "w");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Change the autolock pid file owner to the user
	/////////////////////////////////////////////////

	// Get the user uid using the passwd file
	structPasswd = getpwnam(userName);

	if (structPasswd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to find the user in the passwd file: %s",
			strerror(errno));
		return (PAM_SESSION_ERR);
	}

	// Change the autolock pid file owner
	rCode = chown(PID_AUTOLOCK_FILE, structPasswd->pw_uid, -1);

	if (rCode == -1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to change the owner of the autolock pid file: %sn",
			strerror(errno));
		return (PAM_SESSION_ERR);
	}

	fclose(fd);
	return (EXIT_SUCCESS);
}

int setAutolockPid(int autolockPid)
{
	// This function writes the autolock pid at the 
	// beginning of the autolock pid file.
	//////////////////////////////////////////////////////

	int nbWrote;
	FILE *fd;

	// Open the file
	fd = fopen(PID_AUTOLOCK_FILE, "w+");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Write the autolock pid
	nbWrote = fprintf(fd, "%d\n", autolockPid);

	if (nbWrote < 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to write the autolock pid in autolock pid file (%s)",
			PID_AUTOLOCK_FILE);
		return (EXIT_FAILURE);
	}

	fclose(fd);
	return (EXIT_SUCCESS);
}

int getAutolockPid(int *pid)
{
	// The function retrieves the autolock pid int the autolock pid
	// file and stores it in pid.
	///////////////////////////////////////////////////////////////

	FILE *fd;
	int rCode;
	struct stat statbuf;

	// Open the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "r");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// If the size of the autolock pid file is 0, the autolock
	// doesn't run.
	rCode = stat(PID_AUTOLOCK_FILE, &statbuf);

	if (rCode == -1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to stat the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		fclose(fd);
		return (EXIT_FAILURE);
	}

	else if (statbuf.st_size == 0)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"ERROR: the autolock daemon seems not to be running\n");
		fclose(fd);
		return (EXIT_FAILURE);
	}

	// The autolock pid is the first line of the file
	rCode = fscanf(fd, "%d", pid);

	if (rCode != 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to read the autolock pid in the autolock pid file (%s)",
			PID_AUTOLOCK_FILE);
		fclose(fd);
		return (EXIT_FAILURE);
	}

	fclose(fd);
	return (EXIT_SUCCESS);
}

int getSessionPid(int *pid, int *result, int nr)
{
	// This function retyurn in pid the nr opened session pid
	// registred in the autolock pid file.
	// Result is set to 1 if it is found, set to 0 else
	/////////////////////////////////////////////////////////

	int rCode;
	int i = 0;
	FILE *fd;
	int pidReaded;

	*result = 0;

	// Open the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "r");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Skip the autolock pid
	rCode = fscanf(fd, "%d", &pidReaded);

	if (rCode != 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to skip the autolock pid in the autolock pid file (%s)",
			PID_AUTOLOCK_FILE);
		fclose(fd);
		return (EXIT_FAILURE);
	}

	// Read the first opened session pid
	rCode = fscanf(fd, "%d", &pidReaded);

	while (rCode == 1)
	{
		if (i++ == nr)
		{
			// nr pid found in the list
			*result = 1;
			*pid = pidReaded;
		}

		// Read the next opened session pid
		rCode = fscanf(fd, "%d", &pidReaded);
	}

	fclose(fd);
	return (EXIT_SUCCESS);
}

int listEmpty(int *result)
{
	// This function set result to 1 if the opened session list of 
	// the autolock pid is empty; set result to 0 else.
	//////////////////////////////////////////////////////////////

	int rCode;
	FILE *fd;
	int pidReaded;

	*result = 0;

	// Open the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "r");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Skip the autolock pid
	rCode = fscanf(fd, "%d", &pidReaded);

	if (rCode != 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to skip the autolock pid in the autolock pid file (%s)",
			PID_AUTOLOCK_FILE);
		fclose(fd);
		return (EXIT_FAILURE);
	}

	// Try to read a opened session pid
	rCode = fscanf(fd, "%d", &pidReaded);

	if (rCode != 1)
	{
		// The list is empty
		*result = 1;

		fclose(fd);
		return (EXIT_SUCCESS);
	}

	// The list is not empty  
	fclose(fd);
	return (EXIT_SUCCESS);
}

int putPidInList(int pid)
{
	// This function put pid at the end of the autolock pid file
	////////////////////////////////////////////////////////////

	FILE *fd;
	int nbWrote;

	// Open the autolock pid file an the appen mode
	fd = fopen(PID_AUTOLOCK_FILE, "a+");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Write the pid
	nbWrote = fprintf(fd, "%d\n", pid);

	if (nbWrote < 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to write in the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		fclose(fd);
		return (EXIT_FAILURE);
	}

	fclose(fd);
	return (EXIT_SUCCESS);
}

int deletePidInList(int pid)
{
	// This function delete pid in the autolock pid file.
	// If Pid not in the list nothing is done.
	////////////////////////////////////////////////////

	int rCode;
	FILE *fd;
	char tmpTab[BUF_TMP_SIZE];
	int pidReaded;
	int posInTmpTab = 0;

	memset(tmpTab, '\0', BUF_TMP_SIZE);

	// Open the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "r");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Skip the autolock pid (it's the first bytes)
	// to be positionated at the begining of the session pid list.
	// Copy it in the temporary tab
	rCode = fscanf(fd, "%d", &pidReaded);

	if (rCode != 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to skip the autolock pid in the autolock pid file (%s)",
			PID_AUTOLOCK_FILE);
		fclose(fd);
		return (EXIT_FAILURE);
	}

	posInTmpTab += sprintf(tmpTab + posInTmpTab, "%d\n", pidReaded);

	// Read the first opened session pid
	rCode = fscanf(fd, "%d", &pidReaded);

	while (rCode == 1)
	{
		if (pidReaded == pid)
		{
			// Pid found in the list
			// Don't write the deleted pid in the temporary tab
		}
		else
		{
			// Write the pid in the temporary tab
			posInTmpTab += sprintf(tmpTab + posInTmpTab, "%d\n", pidReaded);
		}

		// Read the next opened session pid
		rCode = fscanf(fd, "%d", &pidReaded);
	}

	fclose(fd);

	// Reopen the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "w");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Copy the temporary tab in the new autolock pid file
	fprintf(fd, "%s", tmpTab);

	fclose(fd);

	return (EXIT_SUCCESS);
}

int pidInList(int pid, int *result)
{
	// This function set result to 1 if pid is registred
	// in the autolock pid file; set result to 0 else.
	////////////////////////////////////////////////////

	int rCode;
	FILE *fd;
	int pidReaded;

	*result = 0;

	// Open the autolock pid file
	fd = fopen(PID_AUTOLOCK_FILE, "r");

	if (fd == NULL)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to open the autolock pid file (%s): %s\n",
			PID_AUTOLOCK_FILE, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Skip the autolock pid (it's the first bytes)
	// to be positionated at the begining of the session pid list.
	rCode = fscanf(fd, "%d", &pidReaded);

	if (rCode != 1)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to skip the autolock pid in the autolock pid file (%s)",
			PID_AUTOLOCK_FILE, strerror(errno));
		fclose(fd);
		return (EXIT_FAILURE);
	}

	// Read the first opened session pid
	rCode = fscanf(fd, "%d", &pidReaded);

	while (rCode == 1)
	{
		if (pidReaded == pid)
		{
			// Pid found in the list
			*result = 1;
			fclose(fd);
			return (EXIT_SUCCESS);
		}

		// Read the next opened session pid
		rCode = fscanf(fd, "%d", &pidReaded);
	}

	// Pid not found in the list
	fclose(fd);
	return (EXIT_SUCCESS);
}

int cleanupList()
{
	// This function check the list in order to verify that all the 
	// registred sessions are still alive
	///////////////////////////////////////////////////////////////

	int i = 0;
	int pid;
	int rCode;
	int result;

	// Get the first opened session pid
	rCode = getSessionPid(&pid, &result, i);

	if (rCode == EXIT_FAILURE)
	{
		pam_log(PAM_SMARTCARD, LOG_ERR,
			"UNABLE to get a session pid from the autolock pid file");
		return (EXIT_FAILURE);
	}

	while (result == 1)
	{
		// To test if the session number pid is still alive,
		// We try to send it the null signal 
		rCode = kill(pid, 0);

		if ((rCode == -1) && (errno == ESRCH))
		{
			// This session doesn't exist
			// Remove it from the list

			pam_log(PAM_SMARTCARD, LOG_ERR,
				"Cleaning pid %d in the autolock pid file", pid);

			if (deletePidInList(pid) == EXIT_FAILURE)
			{
				pam_log(PAM_SMARTCARD, LOG_ERR,
					"Unable to delete a session in the autolock pid file");
				exit(EXIT_FAILURE);
			}
		}
		else
		{
			// Read the next
			i++;
		}

		// Get the next opened session pid
		rCode = getSessionPid(&pid, &result, i);

		if (rCode == EXIT_FAILURE)
		{
			pam_log(PAM_SMARTCARD, LOG_ERR,
				"UNABLE to get an opened session pid from the autolock pid file");
			return (EXIT_FAILURE);
		}
	}

	return (EXIT_SUCCESS);
}

