/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: sst.c,v 1.7 2002/03/11 15:04:58 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>

#include "config.h"
#include "global.h"
#include "smartcard_functions.h"
#include "publickey_db_functions.h"

#define VERSION "1.0"

// This program allows to manage the smartsession cards.
// It's also able to manage the users of the pam_smartcard
// application (add, delete, list the users registred on 
// the system).
// It also manages the administator public key certification.
/////////////////////////////////////////////////////////////

// Internal flags
/////////////////

int twoReaders;					// Two smartcard reader are instaled on the computer

// Internal functions
/////////////////////

// Functions used to get informations from the user
///////////////////////////////////////////////////

// Get or verify user pin code
int get_pin(char *pin);

// Get or verify administrator passphrase
int get_passphrase(char *passphrase);

// Get the public keys destination
int get_publicKeyFile(char *publicKeyFlag, char *publicKeyFile);

// Get or verify the user name
int get_userName(char *userName);

// Functions used to manage the different tasks
///////////////////////////////////////////////

// Print the program version informations
int version();

// Print the program usage informations
int help(char *progName);

// Format a new card for the samrtsession applications  
int format_card(char *passphrase);

// Erase the card
int erase_card(void);

// Get card Directory
int dir_card(void);

// Unlock the card or change the user PIN code on this card
int unlock_card(char *passphrase, char *pin);

// Print the list of the users registred in the database
int list_user(char *publicKeyFile);

// Add a new user on the database
int create_user(char *passphrase, char *publicKeyFile, char *userName);

// Remove an an user from the database
int delete_user(char *publicKeyFile, char *userName);

// Create a new public keys database
int create_db(char *publicKeyFile);

// Create the administrator certification card
int create_certification_card(char *passphrase);

// Install the administrator certification public keys on the computer
int install_certification_key(char *passphrase);


int main(int argc, char **argv)
{
	int rCode;
	char pin[USR_PIN_SIZE + 2] = "";
	char action[BUF_TMP_SIZE] = "";
	char progName[BUF_TMP_SIZE] = "";
	char userName[USER_NAME_MAX_SIZE + 3] = "";
	char passphrase[PASSPHRASE_MAX_LEN + 2] = "";
	char publicKeyFile[DB_NAME_MAX_LEN + 3] = "";

	// Initialize some variables
	twoReaders = 0;

	if (argc < 2)
	{
		// The action option is required
		help(argv[0]);
		return (EXIT_FAILURE);
	}

	// Save the program name and the required action 
	strncpy(progName, argv[0], BUF_TMP_SIZE - 1);
	strncpy(action, argv[1], BUF_TMP_SIZE - 1);

	while (1)
	{
		rCode = getopt(argc, argv, "c:p:d:u:t");

		if (rCode == -1)
			break;

		switch (rCode)
		{
		case 'c':				// User PIN code given in argument
			strncpy(pin, optarg, USR_PIN_SIZE + 1);
			pin[USR_PIN_SIZE + 1] = '0';
			break;

		case 'p':				// Administrator passphrase given in argument
			strncpy(passphrase, optarg, PASSPHRASE_MAX_LEN + 1);
			passphrase[PASSPHRASE_MAX_LEN + 1] = '\0';
			break;

		case 'd':				// Public keys destination given in argument
			strncpy(publicKeyFile, optarg, DB_NAME_MAX_LEN + 1);
			publicKeyFile[DB_NAME_MAX_LEN + 1] = '\0';
			break;

		case 'u':				// User name given in argument
			strncpy(userName, optarg, USER_NAME_MAX_SIZE + 1);
			userName[USER_NAME_MAX_SIZE + 1] = '\0';
			break;

		case 't':				// Two readers are installed on the computer
			twoReaders = 1;
			break;

		case ':':				// Wrong option
			help(progName);
			exit(EXIT_FAILURE);
			break;

		case '?':				// Wrong option
			help(progName);
			exit(EXIT_FAILURE);
			break;
		}
	}

	// It's necessary tu put a lock to be sure to be alone
	// to work on the database.
	if (test_lock() == EXIT_FAILURE)
		exit(EXIT_FAILURE);
	// Unable to put the lock, another smartsession tool process 
	// is running. We have to exit.

	// Next, manage the given action
	////////////////////////////////

	if (strcmp(action, "format_card") == 0)
		rCode = format_card(passphrase);

	else if (strcmp(action, "erase_card") == 0)
		rCode = erase_card();

	else if (strcmp(action, "dir_card") == 0)
		rCode = dir_card();

	else if (strcmp(action, "create_cert_card") == 0)
		rCode = create_certification_card(passphrase);

	else if (strcmp(action, "install_cert_key") == 0)
		rCode = install_certification_key(passphrase);

	else if (strcmp(action, "create_user") == 0)
		rCode = create_user(passphrase, publicKeyFile, userName);

	else if (strcmp(action, "unlock_card") == 0)
		rCode = unlock_card(passphrase, pin);

	else if (strcmp(action, "delete_user") == 0)
		rCode = delete_user(publicKeyFile, userName);

	else if (strcmp(action, "list_user") == 0)
		rCode = list_user(publicKeyFile);

	else if (strcmp(action, "create_db") == 0)
		rCode = create_db(publicKeyFile);

	else if (strcmp(action, "version") == 0)
		rCode = version();

	else
		rCode = help(progName);

	// Now, we can remove the lock file
	unlink(SST_LOCK_FILE);

	return (rCode);
}

int unlock_card(char *passphrase, char *pin)
{
	// This function is used to unlock the card or to change
	// the user PIN code.
	////////////////////////////////////////////////////////

	int rCode;

	// Get or verify the administrator passphrase
	rCode = get_passphrase(passphrase);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Get or verify the new user PIN code
	rCode = get_pin(pin);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Call the pin management function offered by the 
	// smartcard_functions interface.
	rCode = pin_mngmt(passphrase, strlen(passphrase), pin);

	if (rCode == ERROR_CARD_NOT_FORMATED)
	{
		// This error is probably caused by an unformatted card.
		fprintf(stderr, "ERROR: this CARD seems to be NOT FORMATTED!\n");
		return (EXIT_FAILURE);
	}

	if (rCode == ERROR_VERIFY_PIN)
	{
		// The administrator PUC is false, it's probably 
		// caused by a wrong passphrase. 
		fprintf(stderr,
			"ERROR: administrator PUC is FALSE: incorect passphrase!\n");
		return (EXIT_FAILURE);
	}

	if (rCode == EXIT_FAILURE)
	{
		// Other error
		fprintf(stderr, "UNABLE to unlock or change the PIN code!\n");
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}

int format_card(char *passphrase)
{
	// This function is used to format a new card for the 
	// smartsession applications.
	/////////////////////////////////////////////////////

	int rCode;

	// Get or verify the administrator passphrase
	rCode = get_passphrase(passphrase);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Call the format card function offered by 
	// the smartcard_functions interface
	rCode = formatcard(passphrase, strlen(passphrase));

	if (rCode == ERROR_CARD_ALREADY_FORMATED)
	{
		// This error is probably caused by a card already formatted      
		fprintf(stderr, "ERROR: this CARD seems to be ALREADY FORMATTED!\n");
		return (EXIT_FAILURE);
	}

	if (rCode == EXIT_FAILURE)
	{
		// Other error
		fprintf(stderr, "UNABLE to format this card!\n");
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}

int dir_card(void)
{
	// This function is used to erase a card
	/////////////////////////////////////////////////////
	int rCode;

	rCode = dircard();

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	return (EXIT_SUCCESS);
}


int erase_card(void)
{
	// This function is used to erase a card
	/////////////////////////////////////////////////////
	int rCode;

	rCode = erasecard();

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	return (EXIT_SUCCESS);
}

int create_certification_card(char *passphrase)
{
	// This function is used to create the administrator 
	// certification card.
	////////////////////////////////////////////////////

	int rCode;

	// Get or verify the administrator passphrase
	rCode = get_passphrase(passphrase);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Call the generate certification card function offered 
	// by the smartcard_functions interface.
	rCode = generate_certification_part(passphrase, strlen(passphrase));

	if (rCode == ERROR_CARD_ALREADY_FORMATED)
	{
		// This error is probably caused by a card on which a 
		// certification part has already been created.
		fprintf(stderr,
			"ERROR: this CARD seems to be ALREADY A CERTIFICATION CARD\n");
		return (EXIT_FAILURE);
	}

	if (rCode == EXIT_FAILURE)
	{
		// Other error
		printf("UNABLE to create the certification card!\n");
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}

int install_certification_key(char *passphrase)
{
	// This function is used to install the administrator 
	// certification public key on a computer.
	/////////////////////////////////////////////////////

	int rCode;

	// Get or verify the administrator passphrase
	rCode = get_passphrase(passphrase);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Call the install certification card function offered 
	// by the smartcard_functions interface.
	rCode = install_certification_part(passphrase, strlen(passphrase));

	if (rCode == ERROR_CARD_NOT_FORMATED)
	{
		// This error is probably caused by a card on which a certification
		// part hasn't been created.
		fprintf(stderr,
			"ERROR: this CARD seems to be NOT A CERTIFICATION CARD!\n");
		return (EXIT_FAILURE);
	}

	if (rCode == ERROR_VERIFY_PIN)
	{
		// The administrator PUC is false, it's probably 
		// caused by a wrong passphrase.  
		fprintf(stderr,
			"ERROR: administrator PUC is FALSE: incorect passphrase!\n");
		return (EXIT_FAILURE);
	}

	if (rCode == EXIT_FAILURE)
	{
		// Other error
		fprintf(stderr, "UNABLE to install the certification key!\n");
		return (EXIT_FAILURE);
	}
	return (EXIT_SUCCESS);
}

int create_user(char *passphrase, char *publicKeyFlag, char *userName)
{
	// This function add a new user to the specified public
	// keys database.
	///////////////////////////////////////////////////////

	int rCode;
	unsigned char publicKey[RSA_KEY_SIZE_IN_BYTES];
	unsigned char signature[CERT_HASH_PADDING_SIZE];
	char publicKeyFile[DB_NAME_MAX_LEN + 1];

	// Get or verify the administrator passphrase
	rCode = get_passphrase(passphrase);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Get the name of the public keys database
	rCode = get_publicKeyFile(publicKeyFlag, publicKeyFile);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Get or verify the user name
	rCode = get_userName(userName);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Call the perso_card function in order to get the user public
	// key stored in the card and also to store the user name
	// in the card.
	// This function is provided by the smartcard-functions library.
	////////////////////////////////////////////////////////////////

	rCode = perso_card(passphrase, strlen(passphrase), userName,
		publicKeyFile, publicKey);

	if (rCode == ERROR_CARD_NOT_FORMATED)
	{
		// This error is probably caused by an unformatted card.
		fprintf(stderr, "ERROR: this CARD seems to be NOT FORMATTED!\n");
		return (EXIT_FAILURE);
	}

	if (rCode == ERROR_VERIFY_PIN)
	{
		// The administrator PUC is false, it's probably 
		// caused by a wrong passphrase. 
		fprintf(stderr,
			"ERROR: administrator PIN is FALSE: incorect passphrase!\n");
		return (EXIT_FAILURE);
	}

	if (rCode == EXIT_FAILURE)
	{
		// Other error
		fprintf(stderr, "UNABLE to create user %s!\n", userName);
		return (EXIT_FAILURE);
	}

	// Now we have to sign this public key using the administrator
	// certification card.
	//////////////////////////////////////////////////////////////

	if (!twoReaders)
	{
		// Administrator have to insert his card...
		fprintf(stdout, "Please insert the certification card\n");
		fprintf(stdout, "Press ENTER when it's done...\n");

		fflush(stdin);
		fgetc(stdin);

		// Now we can get the signature of the user public key.
		// 0 indicates that the administrator card in the current reader.
		rCode = sign_user_public_key(publicKey, userName, passphrase,
			signature, 0);
	}
	else
	{
		// If there are two readers, the administrator certification card 
		// is supposed to be inserted in the other reader.
		// We can call the signature function with flag set to 1, indicating
		// that the certification card in the other reader.
		rCode = sign_user_public_key(publicKey, userName, passphrase,
			signature, 1);
	}

	if (rCode == ERROR_CARD_NOT_FORMATED)
	{
		// This error is probably caused by a card on which a certification
		// part hasn't been created.
		fprintf(stderr,
			"ERROR: this CARD seems to be NOT A CERTIFICATION CARD!\n");
		return (EXIT_FAILURE);
	}

	if (rCode == ERROR_VERIFY_PIN)
	{
		// The administrator PUC is false, it's probably 
		// caused by a wrong passphrase.  
		fprintf(stderr,
			"ERROR: administrator PUC is FALSE: incorect passphrase!\n");
		return (EXIT_FAILURE);
	}

	if (rCode != EXIT_SUCCESS)
	{
		// Problem during the public key certification
		fprintf(stderr, "UNABLE to create user %s!\n", userName);
		return (EXIT_FAILURE);
	}

	// Now we can add the public key and its signature to the public
	// keys database.
	// flag is set to 0 to indicate that an error must be returned if
	// the user in already registred in the database.
	rCode = publickey_db_add_user(userName, publicKey,
		signature, publicKeyFile, 0);

	if (rCode == ERR_USR_ALREADY_IN_DB)
	{
		// Unable to add this user because it's already registred on the database
		fprintf(stderr,
			"ERROR: user %s ALREADY REGISTRED in the database!\n", userName);
		fprintf(stderr, "Remove this user before add it!\n");
		return (EXIT_FAILURE);
	}

	if (rCode == EXIT_FAILURE)
	{
		// Other error
		fprintf(stderr, "UNABLE to create user %s!\n", userName);
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}

int list_user(char *publicKeyFlag)
{
	// This function is used to print the list of the users 
	// registred on the specified database.
	///////////////////////////////////////////////////////

	int i, size, rCode;
	char **list;
	char publicKeyFile[DB_NAME_MAX_LEN + 1];

	// Get the public keys database name
	rCode = get_publicKeyFile(publicKeyFlag, publicKeyFile);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Call the list user function offered by the 
	// publickey_db_functions interface.
	rCode = publickey_db_list_user(publicKeyFile, &list, &size);

	if (rCode == EXIT_FAILURE)
	{
		// An error occured during the list creation
		fprintf(stderr, "UNABLE to get the list of users!!\n");
		return (EXIT_FAILURE);
	}

	// Print the content of the list and free its elements
	for (i = 0; i < size; i++)
	{
		fprintf(stdout, "%s\n", list[i]);
		free(list[i]);
	}

	free(list);

	return (EXIT_SUCCESS);
}

int delete_user(char *publicKeyFlag, char *userName)
{
	// This function removes the user from the public keys
	// database.
	//////////////////////////////////////////////////////

	int rCode;
	char publicKeyFile[DB_NAME_MAX_LEN + 1];

	// Get the public keys file name
	rCode = get_publicKeyFile(publicKeyFlag, publicKeyFile);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Get or verify the username
	rCode = get_userName(userName);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Call the delete user function offered by the 
	// publickey_db_functions interface.
	rCode = publickey_db_delete_user(userName, publicKeyFile);

	if (rCode == USR_NOT_IN_DB)
	{
		// Unable to remove this user, it's not registred on the database
		fprintf(stderr,
			"ERROR: user %s is NOT REGISTRED in the database\n", userName);
		return (EXIT_FAILURE);
	}

	if (rCode == EXIT_FAILURE)
	{
		// Other error
		fprintf(stderr, "UNABLE to delete this user\n");
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}

int create_db(char *publicKeyFlag)
{
	// This function creates a new database. 
	////////////////////////////////////////

	int rCode;
	char publicKeyFile[DB_NAME_MAX_LEN + 1];

	// Get the public keys file name
	rCode = get_publicKeyFile(publicKeyFlag, publicKeyFile);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Call the database creation function, 
	// 1 for the flag indicates to create a new database.
	rCode = public_key_db_create_db(publicKeyFile, 1);

	if (rCode == EXIT_FAILURE)
	{
		// Error during the database creation
		fprintf(stderr, "UNABLE to create %s database!\n", publicKeyFile);
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}

int help(char *progName)
{
	// Print the program usage informations
	///////////////////////////////////////

	fprintf(stderr, "\n");
	fprintf(stderr, "help: %s action [OPTIONS]\n", progName);
	fprintf(stderr, "\n");
	fprintf(stderr, " action : \n");
	fprintf(stderr, "   format_card [-p passphrase]:\n");
	fprintf(stderr,
		"      format a blank card for the smartsession applications\n");
	fprintf(stderr, "   unlock_card [-p passphrase] [-c code]:\n");
	fprintf(stderr, "      unlock the card or change the user pin\n");
	fprintf(stderr,
		"   create_user [-p passphrase] [-d public_key_destination] [-u usename ] [-t]:\n");
	fprintf(stderr, "      add new user to the system\n");
	fprintf(stderr,
		"   delete_user [-d public_key_destination] [-u usename ]:\n");
	fprintf(stderr, "      remove a user out of the system\n");
	fprintf(stderr, "   list_user [-d public_key_destination]:\n");
	fprintf(stderr, "      list the users of the specified database\n");
	fprintf(stderr, "   create_db [-d public_key_destination]:\n");
	fprintf(stderr, "      create a public keys database\n");
	fprintf(stderr, "   create_cert_card [-p passphrase]:\n");
	fprintf(stderr, "      create the administrator certification card\n");
	fprintf(stderr, "   install_cert_key [-p passphrase]:\n");
	fprintf(stderr,
		"      install the administrator certification public key on the computer\n");
	fprintf(stderr, "\n");
	fprintf(stderr, " OPTIONS\n");
	fprintf(stderr, " -p passphrase : to supply the passphrase\n");
	fprintf(stderr, " -c pin_code : to supply a user pin code\n");
	fprintf(stderr,
		" -d public_key_destination : to supply the destination file\n");
	fprintf(stderr, " -u user_name : to supply the user name\n");
	fprintf(stderr,
		" -t : two reader on the computer, used for the public key certification\n");
	fprintf(stderr, "\n");
	fprintf(stderr, " public_key_destination is in : \n");
	fprintf(stderr,
		"   default : public keys destination is the default database %s\n",
		DEFAULT_DB_NAME);
	fprintf(stderr,
		"   database_name : public keys destination is database_name file\n");
	fprintf(stderr, "\n");

	return (EXIT_FAILURE);
}

int version()
{
	// Print the version informations
	/////////////////////////////////

	fprintf(stderr, "sst " VERSION "\n");
	fprintf(stderr, "  Smart Session Tools by Vincent Guerin\n");
	fprintf(stderr, "  Utilities for management of PAM smart card module\n");
	fprintf(stderr, "  and CFS for smartcard applications\n");
	fprintf(stderr, "    <vincent.guerin@gemplus.com>, 01/09/2000\n");

	return (EXIT_SUCCESS);
}

int get_userName(char *userName)
{
	// This function is used to get the user name or to
	// verify the user name given in argument.
	///////////////////////////////////////////////////

	if (strlen(userName) == 0)
	{
		// No user name given in argument.
		// Ask the administrator for the username.
		//////////////////////////////////////////

		fprintf(stderr, "Enter the user name: ");
		fgets(userName, USER_NAME_MAX_SIZE + 2, stdin);
		userName[strlen(userName) - 1] = '\0';
	}

	// We can verify the given (by administrator or in argument) 
	// username.
	// Only one constraint: a specific size is required.
	////////////////////////////////////////////////////////////

	if (strlen(userName) == 0)
	{
		// No user name given
		fprintf(stderr, "User name MUST BE DEFINED!\n");
		return (EXIT_FAILURE);
	}

	if (strlen(userName) > USER_NAME_MAX_SIZE - 1)
	{
		// User name is too long
		fprintf(stderr,
			"ERROR: user name TOO WIDE, %d characters maximum!\n",
			USER_NAME_MAX_SIZE - 1);
		return (EXIT_FAILURE);
	}

	// pad username with \0
	{
		char *c;

		c = userName + strlen(userName);
		while (c < userName + USER_NAME_MAX_SIZE)
		{
			*(c++) = '\0';
		}
	}

	return (EXIT_SUCCESS);
}

int get_publicKeyFile(char *publicKeyFlag, char *publicKeyFile)
{
	// This function is used to get the public keys destination file
	// according to publickeyFlag.
	// Reminder, it could be default for the default database,
	// another flag will be interpreted as the name of another
	// database. 
	////////////////////////////////////////////////////////////////

	int rCode;

	if (strlen(publicKeyFlag) == 0)
	{
		// No public keys destination given: ask the user for that...
		fprintf(stderr, "Enter the public keys destination: ");
		fgets(publicKeyFlag, DB_NAME_MAX_LEN + 2, stdin);
		publicKeyFlag[strlen(publicKeyFlag) - 1] = '\0';
	}

	if (strlen(publicKeyFlag) > DB_NAME_MAX_LEN)
	{
		// Given destination is too long      
		fprintf(stderr, "ERROR: public keys destination TOO LONG\n");
		fprintf(stderr, "%d characters maximum!\n", DB_NAME_MAX_LEN);
		return (EXIT_FAILURE);
	}

	// Now we can manage the different values of the given (by the 
	// administrator or in argument) publicKeyFlag.
	//////////////////////////////////////////////////////////////

	if (strcmp(publicKeyFlag, "default") == 0)
	{
		// Public keys destination is the default database
		//////////////////////////////////////////////////

		// Create the default database if it doesn't exist yet
		public_key_db_create_db(DEFAULT_DB_NAME, 0);

		// Set the publicKeyFile to the default database
		strcpy(publicKeyFile, DEFAULT_DB_NAME);
	}

	else
	{
		// Another flag is interpreted as a specific database
		///////////////////////////////////////////////////////////

		// Set the publicKeyFile to this other database 
		strcpy(publicKeyFile, publicKeyFlag);

		// Test the existence of this database
		rCode = test_db(publicKeyFile);

		if (rCode == EXIT_FAILURE)
		{
			// If it doesn't exist yet, user must create it himself
			fprintf(stderr, "ERROR: this database DOESN'T EXIST\n");
			fprintf(stderr, "create it before use it!\n");
			return (EXIT_FAILURE);
		}
	}

	return (EXIT_SUCCESS);
}

int get_passphrase(char *passphrase)
{
	// This function is used to get the administrator 
	// passphase or to verify the passphrase given in argument.
	////////////////////////////////////////////////////////////

	char *passphraseReaded;
	char passphrase_1[PASSPHRASE_MAX_LEN + 2];
	char passphrase_2[PASSPHRASE_MAX_LEN + 2];

	if (strlen(passphrase) == 0)
	{
		// Ask the administrator for his passphrase twice.
		// getpass function is used to get this passphrase securely.
		////////////////////////////////////////////////////////////

		passphraseReaded = getpass("Enter your passphrase: ");
		strncpy(passphrase_1, passphraseReaded, PASSPHRASE_MAX_LEN + 1);
		passphrase_1[PASSPHRASE_MAX_LEN + 1] = '\0';

		passphraseReaded = getpass("Re-Enter your passphrase: ");
		strncpy(passphrase_2, passphraseReaded, PASSPHRASE_MAX_LEN + 1);
		passphrase_2[PASSPHRASE_MAX_LEN + 1] = '\0';

		if (strcmp(passphrase_1, passphrase_2) != 0)
		{
			// Obviously the two passphrases must be the same
			fprintf(stderr, "Passphrases MISMATCH, try again!\n");
			return (EXIT_FAILURE);
		}

		strncpy(passphrase, passphrase_1, PASSPHRASE_MAX_LEN + 2);
	}

	// We can verify the given (by administrator or in argument) 
	// passphrase.
	// Only one constraint: the passphrase size must be included
	// between PASSPHRASE_MIN_LEN and PASSPHRASE_MAX_LEN.
	////////////////////////////////////////////////////////////

	if (strlen(passphrase) < PASSPHRASE_MIN_LEN)
	{
		fprintf(stderr,
			"ERROR: passphrase TOO SHORT, at least %d caracters required!\n",
			PASSPHRASE_MIN_LEN);
		return (EXIT_FAILURE);
	}

	if (strlen(passphrase) > PASSPHRASE_MAX_LEN)
	{
		fprintf(stderr,
			"ERROR: passphrase TOO LONG, %d characters maximum!\n",
			PASSPHRASE_MAX_LEN);
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}

int get_pin(char *pin)
{
	// This function is used to get the user pin or to
	// verify the pin given in argument.
	//////////////////////////////////////////////////

	char *pinReaded;
	char pin_1[USR_PIN_SIZE + 2];
	char pin_2[USR_PIN_SIZE + 2];

	if (strlen(pin) == 0)
	{
		// No PIN given in argument.
		// Ask the user for his PIN code twice.
		// getpass function is used to get this PIN securely.
		/////////////////////////////////////////////////////

		pinReaded = getpass("Enter your PIN: ");
		strncpy(pin_1, pinReaded, USR_PIN_SIZE + 1);
		pin_1[USR_PIN_SIZE + 1] = '\0';

		pinReaded = getpass("Re-Enter your PIN: ");
		strncpy(pin_2, pinReaded, USR_PIN_SIZE + 1);
		pin_2[USR_PIN_SIZE + 1] = '\0';

		if (strcmp(pin_1, pin_2) != 0)
		{
			// Obviously the two PINs must be the same
			fprintf(stderr,
				"ERROR: the two PIN codes  MISMATCH, try again!\n");
			return (EXIT_FAILURE);
		}

		strncpy(pin, pin_1, USR_PIN_SIZE + 2);
	}

	// We can verify the given (by user or in argument) PIN.
	// Only one constraint: a specific size is required.
	////////////////////////////////////////////////////////

	if (strlen(pin) != USR_PIN_SIZE)
	{
		fprintf(stderr,
			"ERROR: INCORRECT PIN size, %d characters required!\n",
			USR_PIN_SIZE);
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
